"""
:meta private:
"""
from __future__ import annotations

import warnings
from typing import List, Optional, Union

from pharmpy.deps import sympy
from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.internals.expr.subs import subs
from pharmpy.model import Assignment, Model, NormalDistribution, Parameter, Parameters, Statements

from .blq import get_blq_symb_and_type, get_sd_expr, has_blq_transformation
from .common import _get_unused_parameters_and_rvs, remove_unused_parameters_and_rvs
from .data import get_observations
from .expressions import _create_symbol, create_symbol, get_dv_symbol
from .help_functions import _format_input_list, _get_epsilons
from .parameters import add_population_parameter, fix_parameters, set_initial_estimates


def _preparations(model, y=None):
    stats = model.statements
    # FIXME: Handle other DVs?
    if y is None:
        y = list(model.dependent_variables.keys())[0]
    if not model.statements.find_assignment(y.name):
        raise ValueError(f'Could not find assignment for \'{y}\'')
    f = subs(
        model.statements.find_assignment(y.name).expression,
        {sympy.Symbol(eps): 0 for eps in model.random_variables.epsilons.names},
        simultaneous=True,
    )
    return stats, y, f


def _canonicalize_data_transformation(model, value, dv):
    if value is None:
        value = dv
    else:
        value = parse_expr(value)
        if value.free_symbols != {dv}:
            raise ValueError(
                f"Expression for data transformation must contain the dependent variable "
                f"{dv} and no other variables"
            )
    return value


def remove_error_model(model: Model):
    """Remove error model.

    Parameters
    ----------
    model : Model
        Remove error model for this model

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import remove_error_model, load_example_model
    >>> model = load_example_model("pheno")
    >>> model.statements.find_assignment("Y")
    Y = EPS₁⋅W + F
    >>> model = remove_error_model(model)
    >>> model.statements.find_assignment("Y")
    Y = F

    Warnings
    --------
    Removing the error model will make the model unrunable for some tools.

    """
    stats, y, f = _preparations(model)
    model = model.replace(statements=stats.reassign(y, f))
    model = remove_unused_parameters_and_rvs(model)
    return model.update_source()


def set_additive_error_model(
    model: Model,
    dv: Union[sympy.Symbol, str, int, None] = None,
    data_trans: Optional[Union[str, sympy.Expr]] = None,
    series_terms: int = 2,
):
    r"""Set an additive error model. Initial estimate for new sigma is :math:`(min(DV)/2)²`.

    The error function being applied depends on the data transformation. The table displays
    some examples.

    +------------------------+----------------------------------------+
    | Data transformation    | Additive error                         |
    +========================+========================================+
    | :math:`y`              | :math:`f + \epsilon_1`                 |
    +------------------------+----------------------------------------+
    | :math:`log(y)`         | :math:`\log(f) + \frac{\epsilon_1}{f}` |
    +------------------------+----------------------------------------+

    Parameters
    ----------
    model : Model
        Set error model for this model
    dv : Union[sympy.Symbol, str, int, None]
        Name or DVID of dependent variable. None for the default (first or only)
    data_trans : str or expression
        A data transformation expression or None (default) to use the transformation
        specified by the model. Series expansion will be used for approximation.
    series_terms : int
        Number of terms to use for the series expansion approximation for data
        transformation.

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import set_additive_error_model, load_example_model
    >>> model = load_example_model("pheno")
    >>> model.statements.find_assignment("Y")
    Y = EPS₁⋅W + F
    >>> model = set_additive_error_model(model)
    >>> model.statements.find_assignment("Y")
    Y = F + εₐ

    >>> from pharmpy.modeling import set_additive_error_model, load_example_model
    >>> model = load_example_model("pheno")
    >>> model.statements.find_assignment("Y")
    Y = EPS₁⋅W + F
    >>> model = set_additive_error_model(model, data_trans="log(Y)")
    >>> model.statements.find_assignment("Y")
                 εₐ
        log(F) + ──
    Y =          F

    See Also
    --------
    set_proportional_error_model : Proportional error model
    set_combined_error_model : Combined error model

    """
    dv = get_dv_symbol(model, dv)
    if has_additive_error_model(model, dv):
        return model
    stats, y, f = _preparations(model, dv)
    ruv = create_symbol(model, 'epsilon_a')

    data_trans = _canonicalize_data_transformation(model, data_trans, dv)
    expr = f + ruv

    if data_trans != dv:
        expr = subs(data_trans, {dv: expr}, simultaneous=True).series(ruv, n=series_terms).removeO()

    sigma = create_symbol(model, 'sigma')
    model = add_population_parameter(model, sigma.name, _get_prop_init(model))

    eps = NormalDistribution.create(ruv.name, 'RUV', 0, sigma)

    rvs_new, params_new = _get_unused_parameters_and_rvs(
        stats.reassign(y, expr), model.parameters, model.random_variables + eps
    )

    model = model.replace(
        statements=stats.reassign(y, expr), random_variables=rvs_new, parameters=params_new
    )
    return model.update_source()


def _get_prop_init(model):
    if model.dataset is not None and 'idv' in model.datainfo.types and 'dv' in model.datainfo.types:
        dv_min = get_observations(model).min()
    else:
        dv_min = 0

    if dv_min == 0:
        return 0.01
    else:
        return (dv_min / 2) ** 2


def set_proportional_error_model(
    model: Model,
    dv: Union[sympy.Symbol, str, int, None] = None,
    data_trans: Optional[Union[str, sympy.Expr]] = None,
    zero_protection: bool = True,
):
    r"""Set a proportional error model. Initial estimate for new sigma is 0.09.

    The error function being applied depends on the data transformation.

    +------------------------+----------------------------------------+
    | Data transformation    | Proportional error                     |
    +========================+========================================+
    | :math:`y`              | :math:`f + f \epsilon_1`               |
    +------------------------+----------------------------------------+
    | :math:`log(y)`         | :math:`\log(f) + \epsilon_1`           |
    +------------------------+----------------------------------------+

    Parameters
    ----------
    model : Model
        Set error model for this model
    dv : Union[sympy.Symbol, str, int, None]
        Name or DVID of dependent variable. None for the default (first or only)
    data_trans : str or expression
        A data transformation expression or None (default) to use the transformation
        specified by the model.
    zero_protection : bool
        Set to True to add code protecting from IPRED=0

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = remove_error_model(load_example_model("pheno"))
    >>> model = set_proportional_error_model(model)
    >>> model.statements.after_odes
        A_CENTRAL(t)
        ────────────
    F =      S₁
    W = F
               ⎧2.225e-16  for F = 0
               ⎨
    IPREDADJ = ⎩    F      otherwise
    Y = F + IPREDADJ⋅εₚ
    IPRED = F
    IRES = DV - IPRED
            IRES
            ────
    IWRES =  W

    >>> from pharmpy.modeling import *
    >>> model = remove_error_model(load_example_model("pheno"))
    >>> model = set_proportional_error_model(
    ...     model,
    ...     data_trans="log(Y)"
    ... )
    >>> model.statements.after_odes
        A_CENTRAL(t)
        ────────────
    F =      S₁
    W = F
               ⎧2.225e-16  for F = 0
               ⎨
    IPREDADJ = ⎩    F       otherwise
    Y = εₚ + log(IPREDADJ)
    IPRED = F
    IRES = DV - IPRED
            IRES
            ────
    IWRES =  W

    See Also
    --------
    set_additive_error_model : Additive error model
    set_combined_error_model : Combined error model

    """
    dv = get_dv_symbol(model, dv)
    if has_proportional_error_model(model, dv):
        return model

    stats, y, f = _preparations(model, dv)
    ruv = create_symbol(model, 'epsilon_p')

    data_trans = _canonicalize_data_transformation(model, data_trans, dv)
    ipred = create_symbol(model, 'IPREDADJ') if zero_protection else f

    sigma = create_symbol(model, 'sigma')
    model = add_population_parameter(model, sigma.name, 0.09)

    eps = NormalDistribution.create(ruv.name, 'RUV', 0, sigma)

    f_dummy = sympy.Dummy('x')
    if data_trans == sympy.log(dv):
        error_expr = sympy.log(ipred) + ruv if zero_protection else sympy.log(f_dummy) + ruv
    elif data_trans == dv:
        error_expr = f_dummy + ipred * ruv if zero_protection else f_dummy + f_dummy * ruv
    else:
        raise ValueError(f"Not supported data transformation {data_trans}")

    if has_blq_transformation(model):
        f, stats_new = _get_updated_blq_statements(model, error_expr, f, f_dummy, eps)
    else:
        expr = error_expr.subs({f_dummy: f})
        stats_new = stats.reassign(y, expr)

    if zero_protection:
        guard_expr = sympy.Piecewise((2.225e-16, sympy.Eq(f, 0)), (f, True))
        guard_assignment = Assignment(ipred, guard_expr)
        ind = 0
        # Find first occurrence of IPREDADJ
        for i, s in enumerate(stats_new):
            if ipred in s.free_symbols:
                ind = i
                break
        stats_new = stats_new[0:ind] + guard_assignment + stats_new[ind:]

    rvs_new, params_new = _get_unused_parameters_and_rvs(
        stats_new, model.parameters, model.random_variables + eps
    )

    model = model.replace(statements=stats_new, random_variables=rvs_new, parameters=params_new)

    return model.update_source()


def _get_updated_blq_statements(model, expr_dummy, f, f_dummy, eps_new):
    y = list(model.dependent_variables.keys())[0]
    f_above_lloq = _get_f_above_lloq(model, f)
    expr_above_lloq = expr_dummy.subs({f_dummy: f_above_lloq})
    expr = f.subs({f_above_lloq: expr_above_lloq})
    # FIXME: Make more general
    sd = model.statements.find_assignment('SD')
    sd_new = get_sd_expr(expr_above_lloq, model.random_variables + eps_new, model.parameters)
    stats_new = model.statements.reassign(sd.symbol, sd_new)
    stats_new = stats_new.reassign(y, expr)
    return f_above_lloq, stats_new


def _get_f_above_lloq(model, f):
    blq_symb, _ = get_blq_symb_and_type(model)
    for expr, cond in f.args:
        if blq_symb in cond.free_symbols:
            return expr
    else:
        raise AssertionError('BLQ symbol not found')


def set_combined_error_model(
    model: Model,
    dv: Union[sympy.Symbol, str, int, None] = None,
    data_trans: Optional[Union[str, sympy.Expr]] = None,
):
    r"""Set a combined error model. Initial estimates for new sigmas are :math:`(min(DV)/2)²` for
    proportional and 0.09 for additive.

    The error function being applied depends on the data transformation.

    +------------------------+-----------------------------------------------------+
    | Data transformation    | Combined error                                      |
    +========================+=====================================================+
    | :math:`y`              | :math:`f + f \epsilon_1 + \epsilon_2`               |
    +------------------------+-----------------------------------------------------+
    | :math:`log(y)`         | :math:`\log(f) + \epsilon_1 + \frac{\epsilon_2}{f}` |
    +------------------------+-----------------------------------------------------+

    Parameters
    ----------
    model : Model
        Set error model for this model
    dv : Union[sympy.Symbol, str, int, None]
        Name or DVID of dependent variable. None for the default (first or only)
    data_trans : str or expression
        A data transformation expression or None (default) to use the transformation
        specified by the model.

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = remove_error_model(load_example_model("pheno"))
    >>> model = set_combined_error_model(model)
    >>> model.statements.find_assignment("Y")
    Y = F⋅εₚ + F + εₐ

    >>> from pharmpy.modeling import *
    >>> model = remove_error_model(load_example_model("pheno"))
    >>> model = set_combined_error_model(model, data_trans="log(Y)")
    >>> model.statements.find_assignment("Y")
                     εₐ
       εₚ + log(F) + ──
    Y =              F

    See Also
    --------
    set_additive_error_model : Additive error model
    set_proportional_error_model: Proportional error model

    """
    dv = get_dv_symbol(model, dv)
    if has_combined_error_model(model, dv):
        return model
    stats, y, f = _preparations(model, dv)

    expr = stats.find_assignment(y.name).expression

    ruv_prop = create_symbol(model, 'epsilon_p')
    ruv_add = create_symbol(model, 'epsilon_a')

    eta_ruv = sympy.Symbol('ETA_RV1')
    theta_time = sympy.Symbol('time_varying')

    data_trans = _canonicalize_data_transformation(model, data_trans, dv)

    sigma_prop = create_symbol(model, 'sigma_prop')
    model = add_population_parameter(model, sigma_prop.name, 0.09)
    sigma_add = create_symbol(model, 'sigma_add')
    model = add_population_parameter(model, sigma_add.name, _get_prop_init(model))

    eps_prop = NormalDistribution.create(ruv_prop.name, 'RUV', 0, sigma_prop)
    eps_add = NormalDistribution.create(ruv_add.name, 'RUV', 0, sigma_add)

    # FIXME: Handle other DVs
    dv = list(model.dependent_variables.keys())[0]
    f_dummy = sympy.Dummy('x')
    if data_trans == sympy.log(dv):
        error_expr = sympy.log(f_dummy) + ruv_prop + ruv_add / f_dummy
    elif data_trans == dv:
        # Time varying
        if isinstance(expr, sympy.Piecewise) and not has_blq_transformation(model):
            expr_0 = expr.args[0][0]
            expr_1 = expr.args[1][0]
            cond_0 = expr.args[0][1]
            error_expr = None
            for eps in model.random_variables.epsilons.names:
                expr_0 = subs(expr_0, {sympy.Symbol(eps): ruv_prop}, simultaneous=True)
                expr_1 = subs(expr_1, {sympy.Symbol(eps): ruv_prop}, simultaneous=True)
                if (
                    eta_ruv in model.random_variables.free_symbols
                    and theta_time in model.parameters.symbols
                ):
                    error_expr = sympy.Piecewise(
                        (expr_0 + ruv_add * theta_time * sympy.exp(eta_ruv), cond_0),
                        (expr_1 + ruv_add * sympy.exp(eta_ruv), True),
                    )
                elif (
                    eta_ruv not in model.random_variables.free_symbols
                    and theta_time in model.parameters.symbols
                ):
                    error_expr = sympy.Piecewise(
                        (expr_0 + ruv_add * theta_time, cond_0), (expr_1 + ruv_add, True)
                    )
            assert error_expr is not None
        elif (
            eta_ruv in model.random_variables.free_symbols
            and theta_time not in model.parameters.symbols
        ):
            if has_blq_transformation(model):
                raise ValueError('Currently not supported to change from IIV on RUV model with BLQ')
            error_expr = (
                f_dummy + f_dummy * ruv_prop * sympy.exp(eta_ruv) + ruv_add * sympy.exp(eta_ruv)
            )
        else:
            error_expr = f_dummy + f_dummy * ruv_prop + ruv_add
    else:
        raise ValueError(f"Not supported data transformation {data_trans}")

    if has_blq_transformation(model):
        _, stats_new = _get_updated_blq_statements(
            model, error_expr, f, f_dummy, [eps_prop, eps_add]
        )
    else:
        expr = error_expr.subs({f_dummy: f})
        stats_new = stats.reassign(y, expr)

    rvs_new, params_new = _get_unused_parameters_and_rvs(
        stats_new, model.parameters, model.random_variables + [eps_prop, eps_add]
    )
    model = model.replace(statements=stats_new, random_variables=rvs_new, parameters=params_new)

    return model.update_source()


def has_additive_error_model(model: Model, dv: Union[sympy.Symbol, str, int, None] = None):
    """Check if a model has an additive error model

    Multiple dependent variables are supported. By default the only (in case of one) or the
    first (in case of many) dependent variable is going to be checked.

    Parameters
    ----------
    model : Model
        The model to check
    dv : Union[sympy.Symbol, str, int, None]
        Name or DVID of dependent variable. None for the default (first or only)

    Return
    ------
    bool
        True if the model has an additive error model and False otherwise

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, has_additive_error_model
    >>> model = load_example_model("pheno")
    >>> has_additive_error_model(model)
    False

    See Also
    --------
    has_proportional_error_model : Check if a model has a proportional error model
    has_combined_error_model : Check if a model has a combined error model
    has_weighted_error_model : Check if a model has a weighted error model
    """
    y = get_dv_symbol(model, dv)
    expr = model.statements.error.full_expression(y)
    rvs = model.random_variables.epsilons
    rvs_in_y = {sympy.Symbol(name) for name in rvs.names if sympy.Symbol(name) in expr.free_symbols}
    if len(rvs_in_y) != 1:
        return False
    eps = rvs_in_y.pop()
    return eps not in (expr - eps).simplify().free_symbols


def has_proportional_error_model(model: Model, dv: Union[sympy.Symbol, str, int, None] = None):
    """Check if a model has a proportional error model

    Multiple dependent variables are supported. By default the only (in case of one) or the
    first (in case of many) dependent variable is going to be checked.

    Parameters
    ----------
    model : Model
        The model to check
    dv : Union[sympy.Symbol, str, int, None]
        Name or DVID of dependent variable. None for the default (first or only)

    Return
    ------
    bool
        True if the model has a proportional error model and False otherwise

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, has_proportional_error_model
    >>> model = load_example_model("pheno")
    >>> has_proportional_error_model(model)
    True

    See Also
    --------
    has_additive_error_model : Check if a model has an additive error model
    has_combined_error_model : Check if a model has a combined error model
    has_weighted_error_model : Check if a model has a weighted error model
    """
    y = get_dv_symbol(model, dv)
    expr = model.statements.error.full_expression(y)
    rvs = model.random_variables.epsilons
    rvs_in_y = {sympy.Symbol(name) for name in rvs.names if sympy.Symbol(name) in expr.free_symbols}
    if len(rvs_in_y) != 1:
        return False
    eps = rvs_in_y.pop()

    y_symbs = model.statements.error.find_assignment(y).expression.free_symbols - {eps}
    ipredadj_symb, f_symb = _check_and_get_zero_protect(model.statements.error, y_symbs)
    if ipredadj_symb:
        ipredadj_expr = model.statements.error.full_expression(ipredadj_symb)
        f_expr = model.statements.error.full_expression(f_symb)
        expr = expr.subs({ipredadj_expr: f_expr})

    return eps not in (expr / (1 + eps)).simplify().free_symbols


def _check_and_get_zero_protect(error_sset, y_symbs):
    # FIXME: Support power_on_ruv (currently not needed since there is no has_power_on_ruv)
    if len(y_symbs) > 2:
        return None, None
    f_cand, ipredadj_cand = None, None
    for symb in y_symbs:
        s = error_sset.find_assignment(symb)
        if s and isinstance(s.expression, sympy.Piecewise):
            ipredadj_cand = s
        else:
            f_cand = s
    if ipredadj_cand:
        if ipredadj_cand.expression.free_symbols == {f_cand.symbol}:
            return ipredadj_cand.symbol, f_cand.symbol
    return None, None


def has_combined_error_model(model: Model, dv: Union[sympy.Symbol, str, int, None] = None):
    """Check if a model has a combined additive and proportional error model

    Multiple dependent variables are supported. By default the only (in case of one) or the
    first (in case of many) dependent variable is going to be checked.

    Parameters
    ----------
    model : Model
        The model to check
    dv : Union[sympy.Symbol, str, int, None]
        Name or DVID of dependent variable. None for the default (first or only)

    Return
    ------
    bool
        True if the model has a combined error model and False otherwise

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, has_combined_error_model
    >>> model = load_example_model("pheno")
    >>> has_combined_error_model(model)
    False

    See Also
    --------
    has_additive_error_model : Check if a model has an additive error model
    has_proportional_error_model : Check if a model has a proportional error model
    has_weighted_error_model : Check if a model has a weighted error model
    """
    y = get_dv_symbol(model, dv)
    expr = model.statements.error.full_expression(y)
    rvs = model.random_variables.epsilons
    rvs_in_y = {sympy.Symbol(name) for name in rvs.names if sympy.Symbol(name) in expr.free_symbols}
    if len(rvs_in_y) != 2:
        return False
    eps1 = rvs_in_y.pop()
    eps2 = rvs_in_y.pop()
    canc1 = ((expr - eps1) / (eps2 + 1)).simplify()
    canc2 = ((expr - eps2) / (eps1 + 1)).simplify()
    return (
        eps1 not in canc1.free_symbols
        and eps2 not in canc1.free_symbols
        or eps1 not in canc2.free_symbols
        and eps2 not in canc2.free_symbols
    )


def use_thetas_for_error_stdev(model: Model):
    """Use thetas to estimate standard deviation of error

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, use_thetas_for_error_stdev
    >>> model = load_example_model("pheno")
    >>> model = use_thetas_for_error_stdev(model)
    >>> model.statements.find_assignment("W")
    W = F⋅SD_EPS_1
    >>> model.statements.find_assignment("Y")
    Y = EPS₁⋅W + F

    See also
    --------
    set_weighted_error_model : Encode error model with one epsilon and weight
    """
    rvs = model.random_variables.epsilons
    for eps in rvs:
        sigmas = eps.parameter_names
        if len(sigmas) > 1:
            raise ValueError('use_thetas_for_error_stdev only supports non-correlated sigmas')
        sigma = sigmas[0]

        param = model.parameters[sigma]
        theta_init = param.init**0.5
        model = fix_parameters(model, [sigma])
        model = set_initial_estimates(model, {sigma: 1})

        sdsymb = create_symbol(model, f'SD_{eps.names[0]}')
        model = add_population_parameter(model, sdsymb.name, theta_init, lower=0)
        symb = sympy.Symbol(eps.names[0])
        if has_weighted_error_model(model) and (
            has_additive_error_model(model) or has_proportional_error_model(model)
        ):
            w = get_weighted_error_model_weight(model)
            w_ass = model.statements.find_assignment(w)
            model = model.replace(
                statements=model.statements.reassign(w_ass.symbol, w_ass.expression * sdsymb)
            )
        else:
            model = model.replace(statements=model.statements.subs({symb: sdsymb * symb}))
    return model.update_source()


def set_weighted_error_model(model: Model):
    """Encode error model with one epsilon and W as weight

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, set_weighted_error_model
    >>> model = load_example_model("pheno")
    >>> model = set_weighted_error_model(model)

    See also
    --------
    use_thetas_for_error_stdev : Use thetas to estimate error
    """
    stats, y, f = _preparations(model)
    epsilons = model.random_variables.epsilons
    expr = stats.find_assignment(y.name).expression
    ssum = 0
    q = sympy.Q.real(y)  # Dummy predicate
    for term in expr.args:
        eps = [x for x in term.free_symbols if x.name in epsilons.names]
        if len(eps) > 0:
            eps = eps[0]
            remaining = term / eps
            ssum += remaining**2
            for symb in remaining.free_symbols:
                q &= sympy.Q.positive(symb)
    w = sympy.sqrt(ssum)
    w = sympy.refine(w, q)

    i = _index_of_first_assignment(stats, y)

    model = model.replace(statements=stats[0:i] + Assignment(sympy.Symbol('W'), w) + stats[i:])
    model = model.replace(
        statements=model.statements.reassign(
            y, f + sympy.Symbol('W') * sympy.Symbol(epsilons[0].names[0])
        )
    )
    model = remove_unused_parameters_and_rvs(model)
    return model.update_source()


def has_weighted_error_model(model: Model):
    """Check if a model has a weighted error model

    Parameters
    ----------
    model : Model
        The model to check

    Return
    ------
    bool
        True if the model has a weighted error model and False otherwise

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, has_weighted_error_model
    >>> model = load_example_model("pheno")
    >>> has_weighted_error_model(model)
    True

    See Also
    --------
    has_additive_error_model : Check if a model has an additive error model
    has_combined_error_model : Check if a model has a combined error model
    has_proportional_error_model : Check if a model has a proportional error model
    """
    w = get_weighted_error_model_weight(model)

    if w:
        return True
    return False


def get_weighted_error_model_weight(model: Model):
    # Defines weighted error model as e.g.: Y = F + EPS(1)*W
    stats, y, f = _preparations(model)
    # FIXME: Handle multiple DVs? Handle piecewise?
    y_expr = stats.error.find_assignment(y).expression
    rvs = model.random_variables.epsilons
    rvs_in_y = {
        sympy.Symbol(name) for name in rvs.names if sympy.Symbol(name) in y_expr.free_symbols
    }

    if len(rvs_in_y) > 1:
        return None

    f = y_expr.subs({rv: 0 for rv in rvs_in_y})
    w = None
    eps_expr = {arg for arg in y_expr.args if arg.free_symbols.intersection(rvs_in_y)}

    if len(eps_expr) == 1:
        eps_expr = eps_expr.pop()
        if len(eps_expr.args) == 2 and eps_expr.func is sympy.Mul:
            a, b = eps_expr.args
            w_cand = a if a not in rvs_in_y else b
            if w_cand != f:
                w = w_cand

    return w


def _index_of_first_assignment(statements: Statements, symbol: sympy.Symbol) -> int:
    return next(
        (i for i, s in enumerate(statements) if isinstance(s, Assignment) and s.symbol == symbol)
    )


def set_dtbs_error_model(model: Model, fix_to_log: bool = False):
    """Dynamic transform both sides

    Parameters
    ----------
    model : Model
        Pharmpy model
    fix_to_log : Boolean
        Set to True to fix lambda and zeta to 0, i.e. emulating log-transformed data

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, set_dtbs_error_model
    >>> model = load_example_model("pheno")
    >>> model = set_dtbs_error_model(model)

    """
    model = use_thetas_for_error_stdev(model)
    model = set_weighted_error_model(model)
    stats, y, f = _preparations(model)
    lam = create_symbol(model, 'tbs_lambda')
    zeta = create_symbol(model, 'tbs_zeta')
    if fix_to_log:
        model = add_population_parameter(model, lam.name, 0, fix=True)
        model = add_population_parameter(model, zeta.name, 0, fix=True)
    else:
        model = add_population_parameter(model, lam.name, 1)
        model = add_population_parameter(model, zeta.name, 0.001)

    i = _index_of_first_assignment(stats, sympy.Symbol('W'))

    wass = Assignment(sympy.Symbol('W'), (f**zeta) * sympy.Symbol('W'))
    ipred = sympy.Piecewise(
        ((f**lam - 1) / lam, sympy.And(sympy.Ne(lam, 0), sympy.Ne(f, 0))),
        (sympy.log(f), sympy.And(sympy.Eq(lam, 0), sympy.Ne(f, 0))),
        (-1 / lam, sympy.And(sympy.Eq(lam, 0), sympy.Eq(f, 0))),
        (-1000000000, True),
    )
    ipredass = Assignment(sympy.Symbol('IPRED'), ipred)
    yexpr_ind = stats.find_assignment_index(y.name)
    yexpr = stats[yexpr_ind].subs({f: sympy.Symbol('IPRED')})

    statements = (
        stats[0 : i + 1]
        + wass
        + ipredass
        + stats[i + 1 : yexpr_ind]
        + yexpr
        + stats[yexpr_ind + 1 :]
    )

    obs = model.observation_transformation
    obs = obs.replace(
        y,
        sympy.Piecewise((sympy.log(y), sympy.Eq(lam, 0)), ((y**lam - 1) / lam, sympy.Ne(lam, 0))),
    )
    model = model.replace(observation_transformation=obs, statements=statements)

    return model.update_source()


def set_time_varying_error_model(
    model: Model, cutoff: float, idv: str = 'TIME', dv: Union[sympy.Symbol, str, int, None] = None
):
    """Set a time varying error model per time cutoff

    Parameters
    ----------
    model : Model
        Pharmpy model
    cutoff : float
        A value at the given quantile over idv column
    idv : str
        Time or time after dose, default is Time
    dv : Union[sympy.Symbol, str, int, None]
        Name or DVID of dependent variable. None for the default (first or only)

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, set_time_varying_error_model
    >>> model = load_example_model("pheno")
    >>> model = set_time_varying_error_model(model, cutoff=1.0)
    >>> model.statements.find_assignment("Y")
        ⎧EPS₁⋅W⋅time_varying + F  for TIME < 1.0
        ⎨
    Y = ⎩      EPS₁⋅W + F           otherwise

    """
    dv = get_dv_symbol(model, dv)
    y = model.statements.find_assignment(dv)
    idv = parse_expr(idv)
    theta = create_symbol(model, 'time_varying')
    eps = model.random_variables.epsilons
    expr = sympy.Piecewise(
        (
            subs(
                y.expression,
                {sympy.Symbol(e): sympy.Symbol(e) * theta for e in eps.names},
                simultaneous=True,
            ),
            idv < cutoff,
        ),
        (y.expression, True),
    )
    model = add_population_parameter(model, theta.name, 0.1)
    model = model.replace(statements=model.statements.reassign(y.symbol, expr))

    return model.update_source()


def set_power_on_ruv(
    model: Model,
    list_of_eps: Optional[Union[str, list]] = None,
    dv: Union[sympy.Symbol, str, int, None] = None,
    lower_limit: Optional[float] = 0.01,
    ipred: Optional[Union[str, sympy.Symbol]] = None,
    zero_protection: bool = False,
):
    """Applies a power effect to provided epsilons. If a dependent variable
    is provided, then only said epsilons affecting said variable will be changed.

    Initial estimates for new thetas are 1 if the error
    model is proportional, otherwise they are 0.1.

    NOTE : If no DVs or epsilons are specified, all epsilons with the same name
    will be connected to the same theta. Running the function per DV will give
    each epsilon a specific theta.

    Parameters
    ----------
    model : Model
        Pharmpy model to create block effect on.
    list_of_eps : str or list or None
        Name/names of epsilons to apply power effect. If None, all epsilons will be used.
        None is default.
    dv : Union[sympy.Symbol, str, int, None]
        Name or DVID of dependent variable. None will change the epsilon on all occurences
        regardless of affected dependent variable.
    lower_limit : float or None
        Lower limit of power (theta). None for no limit.
    ipred : Symbol
        Symbol to use as IPRED. Default is to autodetect expression for IPRED.
    zero_protection : bool
        Set to True to add code protecting from IPRED=0

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_power_on_ruv(model)
    >>> model.statements.find_assignment("Y")
              power₁
    Y = EPS₁⋅F       + F

    See also
    --------
    set_iiv_on_ruv

    """
    list_of_eps = _format_input_list(list_of_eps)
    eps = _get_epsilons(model, list_of_eps)
    eps = [e for e in eps]
    dv_symb = get_dv_symbol(model, dv)
    y = model.statements.find_assignment(dv_symb)
    if not y:
        raise ValueError(f'Could not find assignment for \'{dv_symb}\'')

    pset, sset = list(model.parameters), model.statements

    if ipred is None:
        # Extract ipred based on the dv
        ipred = get_ipred(model, dv=dv_symb)
    else:
        ipred = parse_expr(ipred)

    # Assert that the provided epsilons are used for the corresponding DV
    # Else give warning
    if (
        dv is not None
        and list_of_eps is not None
        and any(
            [
                sympy.Symbol(e.names[0])
                not in model.statements.after_odes.full_expression(dv_symb).free_symbols
                for e in eps
            ]
        )
    ):
        warnings.warn(f'Some provided epsilons are not connected to the supplied DV ({dv_symb})')
    elif dv is not None and any(
        [
            sympy.Symbol(e.names[0])
            not in model.statements.after_odes.full_expression(dv_symb).free_symbols
            for e in eps
        ]
    ):
        # Only analyze epsilons connected to the given DV
        eps = [
            e
            for e in eps
            if sympy.Symbol(e.names[0])
            in model.statements.after_odes.full_expression(dv_symb).free_symbols
        ]

    # Check for used DV, not just the first one
    if has_proportional_error_model(model, dv=dv_symb):
        theta_init = 1
    else:
        theta_init = 0.1

    # Find for example W = IPRED
    ipredadj = None
    alternative = None
    for s in sset:
        if isinstance(s, Assignment):
            if s.expression == ipred:
                alternative = s.symbol
                if zero_protection:
                    guard_expr = sympy.Piecewise(
                        (2.225e-307, sympy.Eq(s.expression, 0)), (s.expression, True)
                    )
                    guard_assignment = Assignment(ipred, guard_expr)
                    ind = sset.find_assignment_index('Y')
                    sset = sset[0:ind] + guard_assignment + sset[ind:]
                break
            if isinstance(s.expression, sympy.Piecewise):
                args = s.expression.args
                for expr, cond in args:
                    if expr == ipred:
                        ipredadj = s.symbol
                        break

        if has_blq_transformation(model):
            _, _, f = _preparations(model)
            ipred = _get_f_above_lloq(model, f)

    for e in eps:
        e = e.names[0]
        theta_name = str(
            _create_symbol(
                sset, pset, model.random_variables, model.datainfo, 'power', force_numbering=True
            )
        )
        if lower_limit is None:
            theta = Parameter(theta_name, theta_init)
        else:
            theta = Parameter(theta_name, theta_init, lower=lower_limit)
        pset.append(theta)

        subs_dict = {sympy.Symbol(e) * ipred: sympy.Symbol(e)}  # To avoid getting F*EPS*F**THETA
        if not dv:
            sset = sset.subs(subs_dict)
        else:
            y = y.subs(subs_dict)
            sset = sset.reassign(y.symbol, y.expression)

        if alternative:  # To avoid getting W*EPS*F**THETA
            subs_dict = {sympy.Symbol(e) * alternative: sympy.Symbol(e)}
            if not dv:
                sset = sset.subs(subs_dict)
            else:
                y = y.subs(subs_dict)
                sset = sset.reassign(y.symbol, y.expression)

        if ipredadj:
            subs_dict = {
                sympy.Symbol(e) * ipredadj: ipredadj ** sympy.Symbol(theta.name) * sympy.Symbol(e)
            }
            if not dv:
                sset = sset.subs(subs_dict)
            else:
                y = y.subs(subs_dict)
                sset = sset.reassign(y.symbol, y.expression)
        else:
            subs_dict = {sympy.Symbol(e): ipred ** sympy.Symbol(theta.name) * sympy.Symbol(e)}
            if not dv:
                sset = sset.subs(subs_dict)
            else:
                y = y.subs(subs_dict)
                sset = sset.reassign(y.symbol, y.expression)

        if has_blq_transformation(model):
            # FIXME: Make more general
            y_above_lloq, _ = sset.find_assignment('Y').expression.args[0]
            sd = model.statements.find_assignment('SD')
            sd_new = get_sd_expr(y_above_lloq, model.random_variables, Parameters.create(pset))
            sset = sset.reassign(sd.symbol, sd_new)
    model = model.replace(parameters=Parameters.create(pset), statements=sset)

    return model.update_source()


def get_ipred(model, dv=None):
    # FIXME: Handle other DVs?
    if dv is None:
        dv = list(model.dependent_variables.keys())[0]
    expr = model.statements.after_odes.full_expression(dv)
    ipred = subs(
        expr, {sympy.Symbol(rv): 0 for rv in model.random_variables.names}, simultaneous=True
    )
    for s in model.statements:
        if isinstance(s, Assignment) and s.expression == ipred:
            ipred = s.symbol
            break
    return ipred


def set_iiv_on_ruv(
    model: Model,
    dv: Union[sympy.Symbol, str, int, None] = None,
    list_of_eps: Optional[Union[List[str], str]] = None,
    same_eta: bool = True,
    eta_names: Optional[Union[List[str], str]] = None,
):
    """
    Multiplies epsilons with exponential (new) etas.

    Initial variance for new etas is 0.09.

    Parameters
    ----------
    model : Model
        Pharmpy model to apply IIV on epsilons.
    list_of_eps : str, list
        Name/names of epsilons to multiply with exponential etas. If None, all epsilons will
        be chosen. None is default.
    same_eta : bool
        Boolean of whether all RUVs from input should use the same new ETA or if one ETA
        should be created for each RUV. True is default.
    eta_names : str, list
        Custom names of new etas. Must be equal to the number epsilons or 1 if same eta.
    dv : Union[sympy.Symbol, str, int, None]
        Name or DVID of dependent variable. None for the default (first or only)

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_iiv_on_ruv(model)
    >>> model.statements.find_assignment("Y")
                  ETA_RV1
    Y = EPS₁⋅W⋅ℯ        + F

    See also
    --------
    set_power_on_ruv

    """
    list_of_eps = _format_input_list(list_of_eps)
    eps = _get_epsilons(model, list_of_eps)

    if eta_names and len(eta_names) != len(eps):
        raise ValueError(
            'The number of provided eta names must be equal to the number of epsilons.'
        )

    rvs, pset, sset = model.random_variables, list(model.parameters), model.statements

    if same_eta:
        eta = _create_eta(pset, 1, eta_names)
        rvs = rvs + eta
        eta_dict = {e: eta for e in eps}
    else:
        etas = [_create_eta(pset, i + 1, eta_names) for i in range(len(eps))]
        rvs = rvs + etas
        eta_dict = dict(zip(eps, etas))

    dv_symb = get_dv_symbol(model, dv)
    y = model.statements.find_assignment(dv_symb)

    for e in eps:
        subs_dict = {
            sympy.Symbol(e.names[0]): sympy.Symbol(e.names[0])
            * sympy.exp(sympy.Symbol(eta_dict[e].names[0]))
        }
        # FIXME: This is needed if you e.g. have Y and IPRED, with multiple DVs, how should this be handled?
        if not dv:
            sset = sset.subs(subs_dict)
        else:
            y = y.subs(subs_dict)
            sset = sset.reassign(y.symbol, y.expression)

    model = model.replace(random_variables=rvs, parameters=Parameters.create(pset), statements=sset)
    return model.update_source()


def _create_eta(pset, number, eta_names):
    omega = sympy.Symbol(f'IIV_RUV{number}')
    pset.append(Parameter(str(omega), 0.09))

    if eta_names:
        eta_name = eta_names[number - 1]
    else:
        eta_name = f'ETA_RV{number}'

    eta = NormalDistribution.create(eta_name, 'iiv', 0, omega)
    return eta



from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Iterable, List, Optional, Union

from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.internals.fn.signature import with_same_arguments_as
from pharmpy.internals.fn.type import with_runtime_arguments_type_check
from pharmpy.model import Model
from pharmpy.modeling import add_allometry, get_pk_parameters
from pharmpy.tools import (
    summarize_errors,
    summarize_individuals,
    summarize_individuals_count_table,
    summarize_modelfit_results,
)
from pharmpy.tools.common import ToolResults, update_initial_estimates
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import ModelEntry, Task, Workflow, WorkflowBuilder
from pharmpy.workflows.results import ModelfitResults

if TYPE_CHECKING:
    import sympy
else:
    from pharmpy.deps import sympy


def create_workflow(
    model: Optional[Model] = None,
    results: Optional[ModelfitResults] = None,
    allometric_variable: Union[str, sympy.Expr] = 'WT',
    reference_value: Union[str, int, float, sympy.Expr] = 70,
    parameters: Optional[List[Union[str, sympy.Expr]]] = None,
    initials: Optional[List[Union[int, float]]] = None,
    lower_bounds: Optional[List[Union[int, float]]] = None,
    upper_bounds: Optional[List[Union[int, float]]] = None,
    fixed: bool = True,
):
    """Run allometry tool. For more details, see :ref:`allometry`.

    Parameters
    ----------
    model : Model
        Pharmpy model
    results : ModelfitResults
        Results for model
    allometric_variable : str
        Name of the variable to use for allometric scaling (default is WT)
    reference_value : float
        Reference value for the allometric variable (default is 70)
    parameters : list
        Parameters to apply scaling to (default is all CL, Q and V parameters)
    initials : list
        Initial estimates for the exponents. (default is to use 0.75 for CL and Qs and 1 for Vs)
    lower_bounds : list
        Lower bounds for the exponents. (default is 0 for all parameters)
    upper_bounds : list
        Upper bounds for the exponents. (default is 2 for all parameters)
    fixed : bool
        Should the exponents be fixed or not. (default True)

    Returns
    -------
    AllometryResults
        Allometry tool result object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> from pharmpy.tools import run_allometry, load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> run_allometry(model=model, results=results, allometric_variable='WGT') # doctest: +SKIP

    """

    wb = WorkflowBuilder(name="allometry")
    if model is not None:
        start_task = Task('start_allometry', start, results, model)
    else:
        start_task = Task('start_allometry', start)
    _add_allometry = partial(
        _add_allometry_on_model,
        allometric_variable=allometric_variable,
        reference_value=reference_value,
        parameters=parameters,
        initials=initials,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        fixed=fixed,
    )
    task_add_allometry = Task('add allometry', _add_allometry)
    wb.add_task(task_add_allometry, predecessors=start_task)
    fit_wf = create_fit_workflow(n=1)
    wb.insert_workflow(fit_wf, predecessors=task_add_allometry)
    results_task = Task('results', globals()['results'])
    wb.add_task(results_task, predecessors=[start_task] + fit_wf.output_tasks)
    return Workflow(wb)


def start(modelfit_results, model_or_model_entry):
    if isinstance(model_or_model_entry, ModelEntry):
        return model_or_model_entry
    log = modelfit_results.log if modelfit_results else None
    model_entry = ModelEntry(model_or_model_entry, modelfit_results=modelfit_results, log=log)
    return model_entry


def _add_allometry_on_model(
    input_model_entry,
    allometric_variable,
    reference_value,
    parameters,
    initials,
    lower_bounds,
    upper_bounds,
    fixed,
):
    input_model, input_res = input_model_entry.model, input_model_entry.modelfit_results
    model = update_initial_estimates(input_model, input_res)
    model = add_allometry(
        model,
        allometric_variable=allometric_variable,
        reference_value=reference_value,
        parameters=parameters,
        initials=initials,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        fixed=fixed,
    )

    model = model.replace(name="scaled_model", description="Allometry model")
    model_entry = ModelEntry(model, parent=input_model)
    return model_entry


@with_runtime_arguments_type_check
@with_same_arguments_as(create_workflow)
def validate_input(
    model,
    allometric_variable,
    parameters,
):
    if model is not None:
        validate_allometric_variable(model, allometric_variable)
        validate_parameters(model, parameters)


def _parse_fs(expr: str):
    return map(str, parse_expr(expr).free_symbols)


def validate_allometric_variable(model: Model, allometric_variable: str):
    if not set(_parse_fs(allometric_variable)).issubset(model.datainfo.names):
        raise ValueError(
            f'Invalid `allometric_variable`: got `{allometric_variable}`,'
            f' free symbols must be a subset of {sorted(model.datainfo.names)}.'
        )


def validate_parameters(model: Model, parameters: Optional[Iterable[Union[str, sympy.Expr]]]):
    if parameters is not None:
        allowed_parameters = set(get_pk_parameters(model)).union(
            str(statement.symbol) for statement in model.statements.before_odes
        )
        if not set().union(*map(_parse_fs, parameters)).issubset(allowed_parameters):
            raise ValueError(
                f'Invalid `parameters`: got `{parameters}`,'
                f' must be NULL/None or'
                f' free symbols must be a subset of {sorted(allowed_parameters)}.'
            )


def results(start_model_entry, allometry_model_entry):
    start_model = start_model_entry.model
    start_res = start_model_entry.modelfit_results
    allometry_model = allometry_model_entry.model
    allometry_res = allometry_model_entry.modelfit_results

    allometry_model_failed = allometry_res is None
    best_model = start_model if allometry_model_failed else allometry_model

    summod = summarize_modelfit_results([start_res, allometry_res])
    summod['step'] = [0, 1]
    summods = summod.reset_index().set_index(['step', 'model'])
    suminds = summarize_individuals([start_model, allometry_model], [start_res, allometry_res])
    sumcount = summarize_individuals_count_table(df=suminds)
    sumerrs = summarize_errors([start_res, allometry_res])

    return AllometryResults(
        summary_models=summods,
        summary_individuals=suminds,
        summary_individuals_count=sumcount,
        summary_errors=sumerrs,
        final_model=best_model,
    )


@dataclass(frozen=True)
class AllometryResults(ToolResults):
    pass



import re
import warnings
from typing import TYPE_CHECKING, Dict, List, Optional, Union

import pharmpy.visualization
from pharmpy.internals.expr.units import unit_string
from pharmpy.model import Assignment, Model

from .data import get_observations

if TYPE_CHECKING:
    import altair as alt
    import numpy as np
    import pandas as pd
    import scipy
    import sympy
    import sympy.stats as sympy_stats
else:
    from pharmpy.deps import altair as alt
    from pharmpy.deps import numpy as np
    from pharmpy.deps import pandas as pd
    from pharmpy.deps import scipy, sympy, sympy_stats

norm = scipy.stats.norm


def plot_iofv_vs_iofv(iofv1: pd.Series, iofv2: pd.Series, name1: str, name2: str):
    """Plot individual OFV of two models against each other

    Parameters
    ----------
    iofv1 : pd.Series
        Estimated iOFV of the first model
    iofv2 : pd.Series
        Estimated iOFV of the second model
    name1 : str
        Name of first model
    name2 : str
        Name of second model

    Returns
    -------
    alt.Chart
        Scatterplot

    Examples
    --------

    .. pharmpy-execute::

        from pharmpy.modeling import plot_iofv_vs_iofv
        from pharmpy.tools import load_example_modelfit_results

        res1 = load_example_modelfit_results("pheno")
        res2 = load_example_modelfit_results("pheno_linear")
        plot_iofv_vs_iofv(res1.individual_ofv, res2.individual_ofv, "nonlin", "linear")

    """
    x_label = f'{name1} iOFV'
    y_label = f'{name2} iOFV'
    df = pd.DataFrame(
        {
            x_label: iofv1,
            y_label: iofv2,
        }
    )
    id_name = df.index.name
    df = df.reset_index()
    plot = pharmpy.visualization.scatter_plot_correlation(
        df, x_label, y_label, tooltip_columns=[id_name], title='iOFV vs iOFV'
    )
    return plot


def plot_individual_predictions(
    model: Model, predictions: pd.DataFrame, individuals: Optional[List[int]] = None
):
    """Plot DV and predictions grouped on individuals

    Parameters
    ----------
    model : Model
        Previously run Pharmpy model.
    predictions : pd.DataFrame
        One column for each type of prediction
    individuals : list
        A list of individuals to include. None for all individuals

    Returns
    -------
    alt.Chart
        Plot


    Examples
    --------

    .. pharmpy-execute::

        from pharmpy.modeling import load_example_model, plot_individual_predictions
        from pharmpy.tools import load_example_modelfit_results

        model = load_example_model("pheno")
        res = load_example_modelfit_results("pheno")
        plot_individual_predictions(model, res.predictions, individuals=[1, 2, 3, 4, 5])

    """
    obs = get_observations(model)
    indexcols = predictions.index.names
    idcol = indexcols[0]
    idvcol = indexcols[1]

    data = predictions.join(obs).reset_index()
    data = data.melt(id_vars=indexcols)

    if individuals is not None:
        data = data[data[idcol].isin(individuals)]

    plot = (
        alt.Chart(data)
        .mark_line(point=True)
        .encode(x=idvcol, y='value', color='variable', tooltip=[idvcol, 'value'])
        .facet(f'{idcol}:N', columns=5)
    )
    return plot


def plot_transformed_eta_distributions(
    model: Model,
    parameter_estimates: Union[pd.Series, Dict[str, float]],
    individual_estimates: pd.DataFrame,
):
    """Plot transformed eta distributions for all transformed etas

    Parameters
    ----------
    model : Model
        Previously run Pharmpy model.
    parameter_estimates : Union[pd.Series, Dict[str, float]]
        Parameter estimates of model fit
    individual_estimates : pd.DataFrame
        Individual estimates for etas

    Returns
    -------
    alt.Chart
        Plot
    """
    parameter_estimates = dict(parameter_estimates)
    eta_symbols = {sympy.Symbol(name) for name in model.random_variables.etas.names}

    transformations = []
    for s in model.statements.before_odes:
        if isinstance(s, Assignment):
            m = re.match(r'ETA\w(\d+)', s.symbol.name)
            if m:
                symbols = s.expression.free_symbols
                inter = symbols.intersection(eta_symbols)
                if len(inter) == 1:
                    transformations.append((inter.pop(), s.expression))

    x = np.linspace(-2.0, 2.0, 1000)
    i = 1

    df = pd.DataFrame()

    for eta, expr in transformations:
        var = model.random_variables.etas.get_covariance(eta.name, eta.name).subs(
            parameter_estimates
        )
        subdf = pd.DataFrame({'x': x, 'original': norm.pdf(x, scale=float(var) ** 0.5)})
        rv = sympy_stats.Normal('eta', 0, sympy.sqrt(var))
        expr = expr.subs(parameter_estimates).subs({eta: rv})
        curdens = sympy_stats.density(expr)
        densfn = sympy.lambdify(curdens.variables[0], curdens.expr)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            subdf['boxcox'] = densfn(x)
        subdf = pd.melt(subdf, id_vars=['x'], value_name='density')
        eta_name = f'ETA_{i}'
        ebes = pd.DataFrame(
            {'x': individual_estimates[eta.name], 'density': 0.0, 'variable': 'ebes'}
        )
        subdf = pd.concat([subdf, ebes])

        subdf['eta'] = eta_name
        df = pd.concat([df, subdf])
        i += 1

    single = (
        alt.Chart()
        .mark_area(opacity=0.3)
        .encode(
            x=alt.X('x', axis=alt.Axis(labels=False, ticks=False, title=None)),
            y='density',
            color=alt.Color('variable:N', scale=alt.Scale(domain=['original', 'boxcox'])),
        )
    )
    ticks = (
        alt.Chart()
        .mark_tick()
        .encode(
            x='x',
            y='density',
        )
        .transform_filter(alt.datum.variable == 'ebes')
    )
    layer = alt.layer(single, ticks, data=df)

    facet = layer.facet(facet='eta:N', columns=3)
    return facet


def plot_dv_vs_pred(model: Model, predictions: pd.DataFrame) -> alt.Chart:
    """Plot DV vs PRED

    Parameters
    ----------
    model : Model
        Pharmpy model
    predictions : pd.DataFrame
        DataFrame containing the predictions

    Returns
    -------
    alt.Chart
        Plot

    Examples
    --------

    .. pharmpy-execute::

        from pharmpy.modeling import load_example_model, plot_dv_vs_pred
        from pharmpy.tools import load_example_modelfit_results

        model = load_example_model("pheno")
        res = load_example_modelfit_results("pheno")
        plot_dv_vs_pred(model, res.predictions)

    """
    if 'PRED' in predictions.columns:
        pred = 'PRED'
    else:
        raise ValueError("Cannot find population predictions")
    pred_name = "Population prediction"
    return _dv_vs_anypred(model, predictions, pred, pred_name)


def plot_dv_vs_ipred(model: Model, predictions: pd.DataFrame) -> alt.Chart:
    """Plot DV vs IPRED

    Parameters
    ----------
    model : Model
        Pharmpy model
    predictions : pd.DataFrame
        DataFrame containing the predictions

    Returns
    -------
    alt.Chart
        Plot

    Examples
    --------

    .. pharmpy-execute::

        from pharmpy.modeling import load_example_model, plot_dv_vs_ipred
        from pharmpy.tools import load_example_modelfit_results

        model = load_example_model("pheno")
        res = load_example_modelfit_results("pheno")
        plot_dv_vs_ipred(model, res.predictions)

    """
    if 'CIPREDI' in predictions.columns:
        ipred = 'CIPREDI'
    elif 'IPRED' in predictions.columns:
        ipred = 'IPRED'
    else:
        raise ValueError("Cannot find individual predictions")
    ipred_name = "Individual prediction"
    return _dv_vs_anypred(model, predictions, ipred, ipred_name)


def plot_abs_cwres_vs_ipred(
    model: Model, predictions: pd.DataFrame, residuals: pd.DataFrame
) -> alt.Chart:
    r"""Plot \|CWRES\| vs IPRED

    Parameters
    ----------
    model : Model
        Pharmpy model
    predictions : pd.DataFrame
        DataFrame containing the predictions
    residuals : pd.DataFrame
        DataFrame containing the residuals

    Returns
    -------
    alt.Chart
        Plot

    Examples
    --------

    .. pharmpy-execute::

        from pharmpy.modeling import load_example_model, plot_abs_cwres_vs_ipred
        from pharmpy.tools import load_example_modelfit_results

        model = load_example_model("pheno")
        res = load_example_modelfit_results("pheno")
        plot_abs_cwres_vs_ipred(model, res.predictions, res.residuals)

    """
    if 'CIPREDI' in predictions.columns:
        ipred = 'CIPREDI'
    elif 'IPRED' in predictions.columns:
        ipred = 'IPRED'
    else:
        raise ValueError("Cannot find individual predictions")
    if 'CWRES' not in residuals.columns:
        raise ValueError("CWRES not available in residuals")

    predictions = predictions[[ipred]].reset_index()
    observations = get_observations(model, keep_index=True)
    residuals = abs(residuals[['CWRES']]).set_index(observations.index)
    df = predictions.join(residuals, how='inner').reset_index(drop=True)

    idv = model.datainfo.idv_column.name
    idname = model.datainfo.id_column.name
    dv_unit = model.datainfo.dv_column.unit

    chart = _scatter(
        df,
        x=ipred,
        y='CWRES',
        title="Conditional weighted resuals vs Individual predictions",
        xtitle="Individual predictions",
        ytitle="|Conditional weighted residuals|",
        xunit=dv_unit,
        yunit=dv_unit,
        tooltip=(idname, idv),
    )

    layer = chart + _smooth(chart, ipred, 'CWRES')
    layer = layer.configure_point(size=60)

    return layer


def _dv_vs_anypred(model, predictions, predcol_name, predcol_descr):
    obs = get_observations(model)
    di = model.datainfo
    idv = di.idv_column.name
    dvcol = di.dv_column
    dv = dvcol.name
    dv_unit = dvcol.unit
    idname = di.id_column.name

    predictions = predictions[[predcol_name]]
    df = predictions.join(obs, how='inner').reset_index()

    chart = _grouped_scatter(
        df,
        x=predcol_name,
        y=dv,
        group=idname,
        title=f"Observations vs. {predcol_descr}s",
        xtitle=predcol_descr,
        ytitle="Observation",
        xunit=dv_unit,
        yunit=dv_unit,
        tooltip=(idv,),
    )

    line = _identity_line(
        df[predcol_name].min(), df[predcol_name].max(), df[dv].min(), df[dv].max()
    )

    layer = chart + _smooth(chart, predcol_name, dv) + line
    layer = layer.configure_point(size=60)

    return layer


def plot_cwres_vs_idv(model: Model, residuals: pd.DataFrame) -> alt.Chart:
    """Plot CWRES vs idv

    Parameters
    ----------
    model : Model
        Pharmpy model
    residuals : pd.DataFrame
        DataFrame containing CWRES

    Returns
    -------
    alt.Chart
        Plot

    Examples
    --------

    .. pharmpy-execute::

        from pharmpy.modeling import load_example_model, plot_cwres_vs_idv
        from pharmpy.tools import load_example_modelfit_results

        model = load_example_model("pheno")
        res = load_example_modelfit_results("pheno")
        plot_cwres_vs_idv(model, res.residuals)

    """

    if 'CWRES' not in residuals.columns:
        raise ValueError("CWRES not available in residuals")
    di = model.datainfo
    idv = di.idv_column.name
    idv_unit = di.idv_column.unit
    dv_unit = di.dv_column.unit
    idname = di.id_column.name

    df = residuals[['CWRES']].reset_index()

    chart = _grouped_scatter(
        df,
        x=idv,
        y='CWRES',
        group=idname,
        title="Conditional weighted residuals vs. time",
        xtitle="Time",
        ytitle="CWRES",
        xunit=idv_unit,
        yunit=dv_unit,
        tooltip=(),
    )

    layer = chart + _smooth(chart, idv, 'CWRES') + _zero_line()
    layer = layer.configure_point(size=60)

    return layer


def _smooth(chart, x, y):
    loess_smooth = chart.transform_loess(x, y).mark_line().encode(color=alt.value("#FF0000"))
    return loess_smooth


def _identity_line(xmin, xmax, ymin, ymax):
    line = (
        alt.Chart(
            pd.DataFrame(
                {
                    'var1': [xmin, xmax],
                    'var2': [ymin, ymax],
                }
            )
        )
        .mark_line()
        .encode(x=alt.X('var1'), y=alt.Y('var2'), color=alt.value("#000000"))
    )
    return line


def _zero_line():
    zero_line = alt.Chart().mark_rule().encode(y=alt.datum(0), color=alt.value("#000000"))
    return zero_line


def _title_with_unit(title, unit):
    if unit == 1:
        s = ""
    else:
        s = f" ({unit_string(unit)})"
    return title + s


def _grouped_scatter(df, x, y, group, title, xtitle, ytitle, xunit, yunit, tooltip=()):
    chart = (
        alt.Chart(df)
        .mark_line(point=True, opacity=0.7)
        .encode(
            x=alt.X(x).title(_title_with_unit(xtitle, xunit)),
            y=alt.Y(y).title(_title_with_unit(ytitle, yunit)),
            detail=f"{group}:N",
            tooltip=[x, y, group] + list(tooltip),
        )
        .properties(title=title, width=600, height=300)
        .interactive()
    )
    return chart


def _scatter(df, x, y, title, xtitle, ytitle, xunit, yunit, tooltip=()):
    chart = (
        alt.Chart(df)
        .mark_point(opacity=0.7)
        .encode(
            x=alt.X(x).title(_title_with_unit(xtitle, xunit)),
            y=alt.Y(y).title(_title_with_unit(ytitle, yunit)),
            tooltip=[x, y] + list(tooltip),
        )
        .properties(title=title, width=600, height=300)
        .interactive()
    )
    return chart



class Record:
    """
    Top level class for records.

    Create objects only by using the factory function create_record.
    """

    def __init__(self, name, raw_name, root):
        self.name = name
        self.raw_name = raw_name
        self._root = root

    def replace(self, root):
        return self.__class__(self.name, self.raw_name, root)

    @property
    def root(self):
        """Root of the parse tree"""
        return self._root

    def __str__(self):
        assert self.raw_name is not None
        return self.raw_name + str(self.root)



import math

mathematical_script_capital_n = 'N'  # 𝒩  would be better, but due to bug in
# Qt rendering (i.e. spyder, rstudio) gets misaligned


class Grid:
    def __init__(self, nrows, ncols):
        self.nrows = nrows
        self.ncols = ncols
        self._grid = [[None for x in range(ncols)] for y in range(nrows)]

    def set(self, row, col, value):
        self._grid[row][col] = value

    def _column_widths(self):
        """Calculate the number of characters needed for each column"""
        widths = []
        for col in range(self.ncols):
            maxlen = 0
            for row in range(self.nrows):
                e = self._grid[row][col]
                if e:
                    length = e.minimum_width
                    if length > maxlen:
                        maxlen = length
            widths.append(maxlen)
        return widths

    def _row_height(self, row):
        maxheight = 0
        for col in range(self.ncols):
            e = self._grid[row][col]
            if e:
                n = e.minimum_height
                if n > maxheight:
                    maxheight = n
        return maxheight

    @staticmethod
    def _pad(lines, height):
        n = len(lines)
        if n == height:
            return lines
        else:
            length = len(max(lines, key=len))
            add = height - n
            before = add // 2
            after = add // 2
            if before + after != add:
                before += 1
            s = ' ' * length
            padded = [s] * before + lines + [s] * after
            return padded

    def __str__(self):
        output = ''
        colwidths = self._column_widths()
        for row in range(self.nrows):
            height = self._row_height(row)
            for line in range(height):
                for col in range(self.ncols):
                    width = colwidths[col]
                    e = self._grid[row][col]
                    if e:
                        lines = e.lines(width)
                        padded = Grid._pad(lines, height)
                        output += padded[line]
                    else:
                        output += ' ' * width
                output += '\n'
        return output


def _pad_on_sides(n, char=' '):
    if n <= 0:
        return '', ''
    n_before = n // 2
    n_after = n // 2
    if n_before + n_after != n:
        n_after += 1
    return char * n_before, char * n_after


class Box:
    def __init__(self, text):
        self.text = text.split('\n')

    @property
    def minimum_width(self):
        max_row_len = max(len(s) for s in self.text)
        return max_row_len + 2

    @property
    def minimum_height(self):
        return len(self.text) + 2

    def lines(self, width):
        lines = []
        upper = '┌' + '─' * (width - 2) + '┐'
        lines.append(upper)

        for row in self.text:
            before, after = _pad_on_sides(width - len(row) - 2)
            mid = '│' + before + row + after + '│'
            lines.append(mid)

        lower = '└' + '─' * (width - 2) + '┘'
        lines.append(lower)
        return lines


class Arrow:
    def __init__(self, text, right=True):
        self.text = text
        self.right = right

    @property
    def minimum_width(self):
        return len(self.text) + 3

    @property
    def minimum_height(self):
        return 1

    def lines(self, width):
        before, after = _pad_on_sides(width - self.minimum_width, char='─')
        if self.right:
            return ['─' * 2 + before + self.text + after + '→']
        else:
            return ['←' + before + self.text + after + '─' * 2]


class VerticalArrow:
    def __init__(self, text, down=True):
        self.text = text
        self.down = down

    @property
    def minimum_width(self):
        return len(self.text)

    @property
    def minimum_height(self):
        return 3

    def lines(self, width):
        extra_before, extra_after = _pad_on_sides(width - self.minimum_width)
        n = len(self.text) / 2
        before = extra_before + ' ' * math.floor(n)
        after = extra_after + ' ' * math.ceil(n)
        if self.down:
            return [
                before + '│' + after,
                extra_before + self.text + extra_after,
                before + '↓' + after,
            ]
        else:
            return [
                before + '↑' + after,
                extra_before + self.text + extra_after,
                before + '│' + after,
            ]


class DualVerticalArrows:
    def __init__(self, up_text, down_text):
        self.up_text = up_text
        self.down_text = down_text

    @property
    def minimum_width(self):
        return len(self.up_text) + len(self.down_text) + 1

    @property
    def minimum_height(self):
        return 3

    def lines(self, width):
        nbefore = len(self.up_text) // 2
        nafter = len(self.down_text) // 2
        nmiddle = width - 2 - nbefore - nafter
        before = ' ' * nbefore
        after = ' ' * nafter
        middle = ' ' * nmiddle
        first_line = before + '↑' + middle + '│' + after
        last_line = before + '│' + middle + '↓' + after
        middle_spaces = ' ' * (width - self.minimum_width + 1)
        middle_line = self.up_text + middle_spaces + self.down_text
        return [first_line, middle_line, last_line]


def left_parens(height):
    """Return an array containing each row of a large parenthesis
    used for pretty printing
    """
    a = ['⎧']
    for _ in range(height - 2):
        a.append('⎪')
    a.append('⎩')
    return a


def right_parens(height):
    """Return an array containing each row of a large parenthesis
    used for pretty printing
    """
    a = ['⎫']
    for _ in range(height - 2):
        a.append('⎪')
    a.append('⎭')
    return a


def bracket(a):
    """Append a left bracket for an array of lines"""
    if len(a) == 1:
        return '{' + a[0]
    if len(a) == 2:
        a.append('')
    if (len(a) % 2) == 0:
        upper = len(a) // 2 - 1
    else:
        upper = len(a) // 2
    a[0] = '⎧' + a[0]
    for i in range(1, upper):
        a[i] = '⎪' + a[i]
    a[upper] = '⎨' + a[upper]
    for i in range(upper + 1, len(a) - 1):
        a[i] = '⎪' + a[i]
    a[-1] = '⎩' + a[-1]
    return '\n'.join(a) + '\n'


def _character_to_superscript(ch: str) -> str:
    if ch == "0":
        return "⁰"
    elif ch == "1":
        return "¹"
    elif ch == "2":
        return "²"
    elif ch == "3":
        return "³"
    elif ch == "4":
        return "⁴"
    elif ch == "5":
        return "⁵"
    elif ch == "6":
        return "⁶"
    elif ch == "7":
        return "⁷"
    elif ch == "8":
        return "⁸"
    elif ch == "9":
        return "⁹"
    elif ch == "-":
        return "⁻"
    else:
        return ""


def int_to_superscript(x: int) -> str:
    return ''.join(_character_to_superscript(d) for d in str(x))


def itemize_strings(a):
    # Create an itemization string of a list of strings like:
    # ['a', 'b', 'c'] giving " one of 'a', 'b' or 'c'"
    if len(a) == 1:
        return repr(a[0])
    else:
        quoted = [repr(e) for e in a]
        return f" one of {', '.join(quoted[:-1])} or {quoted[-1]}"



from typing import Callable, Generic, TypeVar

from pharmpy.internals.immutable import Immutable

T = TypeVar('T')


class Task(Generic[T], Immutable):
    """One task in a workflow

    Parameters
    ----------
    name : str
        Name of task
    function : func
        Task function
    task_input : any
        Input arguments to func
    """

    def __init__(self, name: str, function: Callable[..., T], *task_input):
        self._name = name
        self._function = function
        self._task_input: tuple = task_input

    @classmethod
    def create(cls, name: str, function: Callable[..., T], *task_input):
        return cls(name, function, *task_input)

    def replace(self, **kwargs):
        name = kwargs.get("name", self._name)
        function = kwargs.get("function", self._function)
        task_input = kwargs.get("task_input", self._task_input)
        return Task.create(
            name,
            function,
            *task_input,
        )

    @property
    def name(self):
        """Task name"""
        return self._name

    @property
    def function(self):
        """Task function"""
        return self._function

    @property
    def task_input(self):
        """Tuple of static input to function"""
        return self._task_input

    def __repr__(self):
        return self._name



def test_sympy_is_sympy():
    import sympy

    from pharmpy.deps import sympy as sympy2

    assert sympy.Symbol is sympy2.Symbol


def test_sympy_dir():
    import sympy

    from pharmpy.deps import sympy as sympy2

    assert dir(sympy2) == dir(sympy)



from typing import Optional

from pharmpy.internals.immutable import Immutable
from pharmpy.model import Model

from .log import Log
from .results import ModelfitResults


class ModelEntry(Immutable):
    """Model with different modelfit related attributes.

    Object that can connect a model object to results from a modelfit. Attributes that can be connected to the
    model object are the modelfit results, the parent model, and the log.

    A ModelEntry object can be created with only the model object, and results can be attached as a separate step.

    Parameters
    ----------
    model : Model
        Pharmpy model
    parent : Model
        Pharmpy model that is parent to `model` argument (optional)
    modelfit_results : ModelfitResults
        Modelfit results connected to `model` argument (optional)
    log : Log
        Log connected to `model` argument (optional)
    """

    def __init__(
        self,
        model: Model,
        parent: Optional[Model] = None,
        modelfit_results: Optional[ModelfitResults] = None,
        log: Optional[Log] = None,
    ):
        self._model = model
        self._parent = parent
        self._modelfit_results = modelfit_results
        self._log = log

    @classmethod
    def create(
        cls,
        model: Model,
        parent: Optional[Model] = None,
        modelfit_results: Optional[ModelfitResults] = None,
        log: Optional[Log] = None,
    ):
        if parent:
            ModelEntry._canonicalize_parent(model, parent)
        # FIXME: Should we require the same name?
        # if modelfit_results:
        #     ModelEntry._canonicalize_modelfit_results(model, modelfit_results)
        return cls(model=model, parent=parent, modelfit_results=modelfit_results, log=log)

    @staticmethod
    def _canonicalize_parent(model, parent):
        if model.name == parent.name:
            raise ValueError(
                f'Name of `model` and `parent` are the same: `{model.name}`, `{parent.name}`'
            )

    @staticmethod
    def _canonicalize_modelfit_results(model, modelfit_results):
        if model.name != modelfit_results.name:
            raise ValueError(
                f'Name of `model` and `modelfit_results` do not match: `{model.name}`, `{modelfit_results.name}`'
            )

    def attach_results(self, modelfit_results: ModelfitResults, log: Optional[Log] = None):
        """Attaches modelfit results and possible log to ModelEntry objects"""
        if self._modelfit_results:
            raise ValueError('ModelEntry `modelfit_results` attribute already set')
        if self._log:
            raise ValueError('ModelEntry `log` attribute already set')

        return ModelEntry.create(
            self._model, parent=self._parent, modelfit_results=modelfit_results, log=log
        )

    @property
    def model(self):
        """Model object"""
        return self._model

    @property
    def parent(self):
        """Parent model of main model"""
        return self._parent

    @property
    def modelfit_results(self):
        """Modelfit results of main model"""
        return self._modelfit_results

    @property
    def log(self):
        """Log of main model"""
        return self._log



from pharmpy.internals.module.lazy import LazyImport

altair = LazyImport('altair', globals(), 'pharmpy.deps.altair', 'altair')
networkx = LazyImport('networkx', globals(), 'networkx')
numpy = LazyImport('numpy', globals(), 'numpy')
pandas = LazyImport('pandas', globals(), 'pandas')
symengine = LazyImport('symengine', globals(), 'symengine')
sympy = LazyImport('sympy', globals(), 'sympy')
sympy_stats = LazyImport('sympy_stats', globals(), 'sympy.stats')

__all__ = (
    'altair',
    'networkx',
    'numpy',
    'pandas',
    'symengine',
    'sympy',
    'sympy_stats',
)



from pathlib import Path
from typing import Dict, Hashable, Optional, Union

import pytest


@pytest.fixture(scope='session')
def testdata():
    """Test data (root) folder."""
    return Path(__file__).resolve().parent / 'testdata'


@pytest.fixture(scope='session')
def datadir(testdata):
    return testdata / 'nonmem'


@pytest.fixture(scope='session')
def pheno_path(datadir):
    return datadir / 'pheno_real.mod'


@pytest.fixture(scope='session')
def pheno(load_model_for_test, pheno_path):
    return load_model_for_test(pheno_path)


@pytest.fixture(scope='session')
def load_model_for_test(tmp_path_factory):
    from pharmpy.model import Model

    _cache: Dict[Hashable, Model] = {}

    def _load(given_path: Union[str, Path]) -> Model:
        # TODO: Cache based on file contents instead.

        def _parse_model():
            model = Model.parse_model(given_path)
            try:
                model.dataset  # NOTE: Force parsing of dataset
            except FileNotFoundError:
                pass  # NOTE: The error will resurface later if needed
            return model

        basetemp = tmp_path_factory.getbasetemp().resolve()

        resolved_path = Path(given_path).resolve()

        try:
            # NOTE: This skips caching when we are reading from a temporary
            # directory.
            resolved_path.relative_to(basetemp)
            return _parse_model()
        except ValueError:
            # NOTE: This is raised when resolved_path is not descendant of
            # basetemp. With Python >= 3.9 we could use is_relative_to instead.
            pass

        from pharmpy.tools.external.nonmem import conf

        key = (str(conf), str(resolved_path))

        if key not in _cache:
            _cache[key] = _parse_model()

        return _cache[key]

    return _load


@pytest.fixture(scope='session')
def load_example_model_for_test():
    from pharmpy.model import Model
    from pharmpy.modeling import load_example_model

    _cache: Dict[Hashable, Model] = {}

    def _load(given_name: str) -> Model:
        def _parse_model():
            return load_example_model(given_name)

        from pharmpy.tools.external.nonmem import conf

        key = (str(conf), given_name)

        if key not in _cache:
            _cache[key] = _parse_model()

        return _cache[key]

    return _load


@pytest.fixture(scope='session')
def create_model_for_test(load_example_model_for_test):
    from pharmpy.model import Model

    def _create(code: str, dataset: Optional[str] = None) -> Model:
        model = Model.parse_model_from_string(code)
        datapath = model.datainfo.path
        if dataset is not None:
            # NOTE: This yields a copy of the dataset through Model#copy
            model = model.replace(
                dataset=load_example_model_for_test(dataset).dataset,
                datainfo=model.datainfo.replace(path=datapath),
            )
        return model

    return _create



import shutil

from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import Model
from pharmpy.modeling import (  # convert_model,; create_basic_pk_model,
    remove_parameter_uncertainty_step,
    transform_blq,
)
from pharmpy.tools import (  # fit,; run_structsearch,
    read_modelfit_results,
    retrieve_final_model,
    run_tool,
)

# from pharmpy.tools.structsearch.pkpd import create_pk_model


# def test_ruvsearch_dv(tmp_path, load_model_for_test, testdata):
#    with chdir(tmp_path):
#        dataset = testdata / "nonmem/pheno_pd.csv"
#        model = create_basic_pk_model("iv", dataset_path=dataset)
#        model = convert_model(model, 'nonmem')

# Fit pk model
#        pk_model = create_pk_model(model)
#        pk_res = fit(pk_model)

# structsearch
#        res_struct = run_structsearch(type='pkpd', results=pk_res, model=model)
#        model = res_struct.models[1]

# ruvsearch
#        res_ruv = run_tool('ruvsearch', model=model, results=model.modelfit_results, dv=2)
#        assert res_ruv


def test_ruvsearch(tmp_path, testdata):
    with chdir(tmp_path):
        for path in (testdata / 'nonmem' / 'ruvsearch').glob('mox3.*'):
            shutil.copy2(path, tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'ruvsearch' / 'moxo_simulated_resmod.csv', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'ruvsearch' / 'mytab', tmp_path)

        model = Model.parse_model('mox3.mod')
        results = read_modelfit_results('mox3.mod')
        model = remove_parameter_uncertainty_step(model)
        res = run_tool('ruvsearch', model=model, results=results, groups=4, p_value=0.05, skip=[])
        iteration = [1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3]
        best_model = retrieve_final_model(res)
        assert (res.cwres_models.index.get_level_values('iteration') == iteration).all()
        assert best_model.model_code.split('\n')[12] == 'IF (TAD.LT.6.08) THEN'
        assert (
            best_model.model_code.split('\n')[13]
            == '    Y = A(2)/VC + A(2)*EPS(1)*THETA(4)*EXP(ETA_RV1)/VC'
        )
        assert (
            best_model.model_code.split('\n')[15] == '    Y = A(2)/VC + A(2)*EPS(1)*EXP(ETA_RV1)/VC'
        )
        assert best_model.model_code.split('\n')[20] == '$THETA  1.15573 ; time_varying'
        assert best_model.model_code.split('\n')[26] == '$OMEGA  0.0396751 ; IIV_RUV1'


def test_ruvsearch_blq(tmp_path, testdata):
    with chdir(tmp_path):
        for path in (testdata / 'nonmem' / 'ruvsearch').glob('mox3.*'):
            shutil.copy2(path, tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'ruvsearch' / 'moxo_simulated_resmod.csv', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'ruvsearch' / 'mytab', tmp_path)

        # Introduce 0 in CWRES to mimic rows BLQ
        with open('mytab') as f:
            mytab_new = f.read().replace('-2.4366E+00', '0.0000E+00')

        with open('mytab', 'w') as f:
            f.write(mytab_new)

        model = Model.parse_model('mox3.mod')
        results = read_modelfit_results('mox3.mod')

        model = transform_blq(model, method='m4', lloq=0.05)

        assert len(results.residuals.loc[results.residuals['CWRES'] == 0]) > 0

        res = run_tool(
            'ruvsearch',
            model=model,
            results=results,
            skip=['IIV_on_RUV', 'time_varying'],
            max_iter=1,
        )

        assert len(res.cwres_models) > 1



import json
import os
import subprocess
import sys
import uuid
import warnings
from pathlib import Path
from typing import Optional, Union

import pharmpy.model
from pharmpy.deps import pandas as pd
from pharmpy.internals.code_generator import CodeGenerator
from pharmpy.model.external.nlmixr import convert_model
from pharmpy.model.external.nlmixr.model import add_evid
from pharmpy.modeling import (
    append_estimation_step_options,
    get_sigmas,
    get_thetas,
    set_evaluation_step,
    update_inits,
    write_csv,
)
from pharmpy.tools import fit
from pharmpy.workflows import ModelEntry, default_tool_database
from pharmpy.workflows.log import Log
from pharmpy.workflows.results import ModelfitResults


def execute_model(model_entry, db, evaluate=False, path=None):
    assert isinstance(model_entry, ModelEntry)
    model = model_entry.model

    if evaluate:
        if [s.evaluation for s in model.estimation_steps._steps][0] is False:
            model = set_evaluation_step(model)

    database = db.model_database
    model = convert_model(model)
    if path is None:
        path = Path.cwd() / f'nlmixr_run_{model.name}-{uuid.uuid1()}'
    model = model.replace(internals=model.internals.replace(path=path))
    meta = path / '.pharmpy'
    meta.mkdir(parents=True, exist_ok=True)
    if model.datainfo.path is not None:
        model = model.replace(datainfo=model.datainfo.replace(path=None))
    write_csv(model, path=path)
    model = model.replace(datainfo=model.datainfo.replace(path=path))

    dataname = f'{model.name}.csv'
    if sys.platform == 'win32':
        dataset_path = f"{path / dataname}".replace("\\", "\\\\")
    else:
        dataset_path = f"{path / dataname}"
    pre = f'library(nlmixr2)\n\ndataset <- read.csv("{dataset_path}")\n'

    if "fix_eta" in model.estimation_steps[0].tool_options:
        fix_eta_filename = "fix_eta.csv"
        if sys.platform == 'win32':
            fix_eta_path = f"{path / fix_eta_filename}".replace("\\", "\\\\")
        else:
            fix_eta_path = f"{path / fix_eta_filename}"
        pre += f'etas <- as.matrix(read.csv("{fix_eta_path}"))'
    pre += "\n"

    code = pre + model.model_code
    cg = CodeGenerator()
    cg.add('ofv <- fit$objDf$OBJF')
    cg.add('thetas <- as.data.frame(fit$theta)')
    cg.add('omega <- fit$omega')
    cg.add('sigma <- as.data.frame(fit$theta)')
    cg.add('log_likelihood <- fit$objDf$`Log-likelihood`')
    cg.add('runtime_total <- sum(fit$time)')
    cg.add('pred <- as.data.frame(fit[c("ID", "TIME", "PRED", "IPRED")])')

    if sys.platform == 'win32':
        p = f"{path / model.name}.RDATA".replace("\\", "\\\\")
    else:
        p = f"{path / model.name}.RDATA"
    cg.add(f'save(file="{p}",ofv, thetas, omega, sigma, log_likelihood, runtime_total, pred)')
    code += f'\n{str(cg)}'
    with open(path / f'{model.name}.R', 'w') as fh:
        fh.write(code)

    from pharmpy.tools.external.nlmixr import conf

    rpath = conf.rpath / 'bin' / 'Rscript'

    newenv = os.environ
    # Reset environment variables incase started from R
    # and calling other R version.
    newenv['R_LIBS_USERS'] = ''
    newenv['R_LIBS_SITE'] = ''

    stdout = path / 'stdout'
    stderr = path / 'stderr'

    args = [str(rpath), str(path / (model.name + '.R'))]

    with open(stdout, "wb") as out, open(stderr, "wb") as err:
        result = subprocess.run(args, stdin=subprocess.DEVNULL, stderr=err, stdout=out, env=newenv)

    rdata_path = path / f'{model.name}.RDATA'

    metadata = {
        'plugin': 'nlmixr',
        'path': str(path),
    }

    plugin = {
        'rpath': str(rpath),
        'commands': [
            {
                'args': args,
                'returncode': result.returncode,
                'stdout': 'stdout',
                'stderr': 'stderr',
            }
        ],
    }

    with database.transaction(model) as txn:
        txn.store_local_file(path / f'{model.name}.R')
        txn.store_local_file(rdata_path)

        txn.store_local_file(stdout)
        txn.store_local_file(stderr)

        txn.store_local_file(model.datainfo.path)

        plugin_path = path / 'nlmixr.json'
        with open(plugin_path, 'w') as f:
            json.dump(plugin, f, indent=2)

        txn.store_local_file(plugin_path)

        txn.store_metadata(metadata)
        txn.store_modelfit_results()

    res = parse_modelfit_results(model, path)
    log = res.log if res else None
    model_entry = model_entry.attach_results(modelfit_results=res, log=log)

    return model_entry


def verification(
    model: pharmpy.model.Model,
    modelfit_results: Optional[ModelfitResults] = None,
    error: float = 10**-3,
    return_comp: bool = False,
    return_stat: bool = False,
    fix_eta: bool = True,
    force_ipred: bool = False,
    force_pred: bool = False,
    ignore_print=False,
) -> Union[bool, pd.DataFrame]:
    """
    Verify that a model inputet in NONMEM format can be correctly translated to
    nlmixr as well as verify that the predictions of the two models are the same
    given a user specified error margin (defailt is 0.001).

    Comparison will be done on PRED values as default unless only IPRED values
    are present or if force_ipred is set to True. Can also force to use pred values
    which will cause an error if only ipred values are present.

    Parameters
    ----------
    model : pharmpy.model.Model
        pharmpy Model object in NONMEM format.
    modelfit_results : ModelfitResults
        pharmpy ModelfitResults of input model.
    error : float, optional
        Allowed error margins for predictions. The default is 10**-3.
    return_comp : bool, optional
        Choose to return table of predictions. The default is False.
    fix_eta : bool, optional
        Decide if NONMEM estimated ETAs are to be used. The default is True.
    force_ipred : bool, optional
        Force to use IPRED for calculating differences instead of PRED. The default is False.
    force_pred : bool, optional
        Force to use PRED for calculating differences. The default is False.

    Returns
    -------
    Union[bool, pd.Dataframe]
        If return_comp = True, return a table of comparisons and differences in
        predictions instead of a boolean indicating if they are the same or not

    """

    nonmem_model = model

    try:
        # Save results from the NONMEM model
        if modelfit_results is None:
            if not ignore_print:
                print_step("Calculating NONMEM predictions... (this might take a while)")
            nonmem_res = fit(nonmem_model)
        else:
            if modelfit_results.predictions is None:
                if not ignore_print:
                    print_step("Calculating NONMEM predictions... (this might take a while)")
                nonmem_res = fit(nonmem_model)
            else:
                nonmem_res = modelfit_results
    except Exception:
        raise Exception("Nonmem model could not be fitted")

    # Set a tool option to fix theta values when running nlmixr
    if fix_eta:
        nonmem_model = fixate_eta(nonmem_model)

    # Check that evaluation step is set to True
    if [s.evaluation for s in nonmem_model.estimation_steps._steps][0] is False:
        nonmem_model = set_evaluation_step(nonmem_model)

    # Update the NONMEM model with new estimates
    # and convert to nlmixr
    if not ignore_print:
        print_step("Converting NONMEM model to nlmixr2...")
    try:
        nlmixr_model = convert_model(update_inits(nonmem_model, nonmem_res.parameter_estimates))
    except Exception:
        raise Exception("Could not convert model to nlmixr2")

    # Execute the nlmixr model
    db = default_tool_database(toolname="comparison")
    if not ignore_print:
        print_step("Executing nlmixr2 model... (this might take a while)")
    path = Path.cwd() / f'nlmixr_run_{model.name}-{uuid.uuid1()}'
    meta = path / '.pharmpy'
    meta.mkdir(parents=True, exist_ok=True)
    write_fix_eta(nonmem_res, path=path)
    try:
        # FIXME : use fit() instead and incorporate write_fix_eta
        nlmixr_model_entry = ModelEntry.create(nlmixr_model, modelfit_results=None)
        nlmixr_model_entry = execute_model(nlmixr_model_entry, db, path=path)
    except Exception:
        raise Exception("nlmixr2 model could not be fitted")

    # Combine the two based on ID and time
    if not ignore_print:
        print_step("Creating result comparison table...")
    combined_result = compare_models(
        nonmem_model,
        nonmem_res,
        nlmixr_model,
        nlmixr_model_entry.modelfit_results,
        error=error,
        force_ipred=force_ipred,
        force_pred=force_pred,
        return_stat=return_stat,
        ignore_print=ignore_print,
    )

    combined_result.to_csv(db.path / "comparison.csv", index=False)

    if not ignore_print:
        print_step("DONE")
    if return_comp is True or return_stat is True:
        return combined_result
    else:
        if all(combined_result["PASS/FAIL"] == "PASS"):
            return True
        else:
            return False


def compare_models(
    model_1,
    model_1_res,
    model_2,
    model_2_res,
    error=10**-3,
    force_ipred=False,
    force_pred=False,
    ignore_print=False,
    return_stat=False,
):
    assert model_1_res.predictions is not None
    assert model_2_res.predictions is not None

    mod1 = model_1
    mod1_type = str(type(mod1)).split(".")[3]
    mod2 = model_2
    mod2_type = str(type(mod2)).split(".")[3]

    nm_to_r = False
    if (mod1_type == "nonmem" and mod2_type != "nonmem") or (
        mod2_type == "nonmem" and mod1_type != "nonmem"
    ):
        nm_to_r = True

    mod1 = mod1.replace(dataset=mod1.dataset.reset_index())
    mod2 = mod2.replace(dataset=mod2.dataset.reset_index())

    if "EVID" not in mod1.dataset.columns:
        mod1 = add_evid(mod1)

    if "EVID" not in mod2.dataset.columns:
        mod2 = add_evid(mod2)

    if nm_to_r:
        if mod1_type == "nonmem":
            predictions = model_1_res.predictions.reset_index()
            predictions = predictions.drop(
                mod1.dataset[~mod1.dataset["EVID"].isin([0, 2])].index.to_list()
            )
            predictions = predictions.set_index(["ID", "TIME"])
            mod1_res_pred = ModelfitResults(predictions=predictions)
            mod2_res_pred = model_2_res

            dataset = mod1.dataset
            dv_var = "DV"
            dv = dataset.drop(mod1.dataset[~mod1.dataset["EVID"].isin([0, 2])].index.to_list())[
                dv_var
            ]
        if mod2_type == "nonmem":
            predictions = model_2_res.predictions.reset_index()
            predictions = predictions.drop(
                mod2.dataset[~mod2.dataset["EVID"].isin([0, 2])].index.to_list()
            )
            predictions = predictions.set_index(["ID", "TIME"])

            mod1_res_pred = model_1_res
            mod2_res_pred = ModelfitResults(predictions=predictions)

            dataset = mod2.dataset
            dv_var = "DV"
            dv = dataset.drop(mod2.dataset[~mod2.dataset["EVID"].isin([0, 2])].index.to_list())[
                dv_var
            ]
    else:
        dataset = mod1.dataset
        dv_var = "DV"
        dv = dataset.drop(mod1.dataset[~mod1.dataset["EVID"].isin([0, 2])].index.to_list())[dv_var]

    dv = dv.reset_index(drop=True)

    mod1_results = mod1_res_pred.predictions.copy()

    mod2_results = mod2_res_pred.predictions.copy()

    pred = False
    ipred = False
    # ---
    if force_pred:
        if "PRED" in mod1_results.columns:
            pred = True
            p = "PRED"
            assert p in mod2_results.columns
            mod1_results.rename(columns={p: f'PRED_{mod1_type}'}, inplace=True)
            mod2_results.rename(columns={p: f'PRED_{mod2_type}'}, inplace=True)
        else:
            print("No PRED column found")
            return None
    elif force_ipred:
        if "IPRED" in mod1_results.columns:
            p = "IPRED"
            ipred = True
            assert p in mod2_results.columns
            mod1_results.rename(columns={p: f'{p}_{mod1_type}'}, inplace=True)
            mod2_results.rename(columns={p: f'IPRED_{mod2_type}'}, inplace=True)
        elif "CIPREDI" in mod1_results.columns:
            p = "CIPREDI"
            ipred = True
            assert p in mod2_results.columns
            mod1_results.rename(columns={p: f'{p}_{mod1_type}'}, inplace=True)
            mod2_results.rename(columns={p: f'IPRED_{mod2_type}'}, inplace=True)
        else:
            print("No IPRED (or CIPRED) column found")
            return None
    else:
        if "PRED" in mod1_results.columns:
            p = "PRED"
            pred = True
            assert p in mod2_results.columns
            mod1_results.rename(columns={p: f'PRED_{mod1_type}'}, inplace=True)
            mod2_results.rename(columns={p: f'PRED_{mod2_type}'}, inplace=True)
        elif "IPRED" in mod1_results.columns:
            p = "IPRED"
            ipred = True
            mod1_results.rename(columns={p: f'{p}_{mod1_type}'}, inplace=True)
            assert p in mod2_results.columns
            mod2_results.rename(columns={p: f'{p}_{mod2_type}'}, inplace=True)
        elif "CIPREDI" in mod1_results.columns:
            p = "CIPREDI"
            ipred = True
            mod1_results.rename(columns={p: f'IPRED_{mod1_type}'}, inplace=True)
            if p not in mod2_results.columns:
                mod2_results.rename(columns={"IPRED": f'IPRED_{mod2_type}'}, inplace=True)
            else:
                mod2_results.rename(columns={p: f'IPRED_{mod2_type}'}, inplace=True)

    if not (pred or ipred):
        print("No comparable prediction value was found. Please use 'PRED' or 'IPRED")
        return False

    combined_result = mod1_results
    if pred:
        combined_result[f'PRED_{mod2_type}'] = mod2_results[f'PRED_{mod2_type}'].to_list()
        # Add difference between the models
        combined_result['PRED_DIFF'] = abs(
            combined_result[f'PRED_{mod1_type}'] - combined_result[f'PRED_{mod2_type}']
        )
    if ipred:
        combined_result[f'IPRED_{mod2_type}'] = mod2_results[f'IPRED_{mod2_type}'].to_list()
        combined_result['IPRED_DIFF'] = abs(
            combined_result[f'IPRED_{mod1_type}'] - combined_result[f'IPRED_{mod2_type}']
        )

    combined_result["DV"] = dv.values

    combined_result["PASS/FAIL"] = "PASS"
    if not ignore_print:
        print("Differences in population predicted values")
    if (pred and ipred) or (pred and not ipred):
        if force_ipred:
            if not ignore_print:
                print("Using IPRED values for final comparison")
            final = "IPRED"
        else:
            if not ignore_print:
                print("Using PRED values for final comparison")
            final = "PRED"
    elif ipred and not pred:
        if force_ipred:
            if not ignore_print:
                print("Using IPRED values for final comparison")
        else:
            if not ignore_print:
                print("Using IPRED values instead")
        final = "IPRED"

    combined_result.loc[combined_result[f'{final}_DIFF'] > error, "PASS/FAIL"] = "FAIL"
    if not ignore_print:
        print(
            combined_result[f'{final}_DIFF'].describe()[["min", "mean", "75%", "max"]].to_string(),
            end="\n\n",
        )

    if return_stat:
        return combined_result[f'{final}_DIFF'].describe()[["min", "mean", "75%", "max"]]
    else:
        return combined_result


def print_step(s: str) -> None:
    """
    Print step currently being performed. Used during verification

    Parameters
    ----------
    s : str
        Information to print.

    See also
    --------
    verification : verify conversion of model to nlmixr2

    """
    print("***** ", s, " *****")


def fixate_eta(model: pharmpy.model.Model) -> pharmpy.model.Model:
    """
    Used during verification to give information to model to fixate etas to
    NONMEM estimates. Add the information to the models tool_options

    Parameters
    ----------
    model : pharmpy.model.Model
        An nlmixr2 pharmpy model object to fixate etas for.

    Returns
    -------
    model : TYPE
        Model with modified tool options to fixate etas during verification.

    See also
    --------
    verification : verify conversion of model to nlmixr2

    """
    opts = {"fix_eta": True}
    model = append_estimation_step_options(model, tool_options=opts, idx=0)
    return model


def write_fix_eta(modelfit_results: ModelfitResults, path=None, force=True) -> str:
    """
    Writes ETAs to be fixated during verification to a csv file to be read by
    nlmixr2

    Parameters
    ----------
    model : pharmpy.model.Model
        A pharmpy model object.
    path : TYPE, optional
        Path to write csv file to. The default is None.
    force : TYPE, optional
        Force overwrite the file if exist. The default is True.

    Raises
    ------
    FileExistsError
        If csv file exist and force is False, raise error.

    Returns
    -------
    str
        Return the path to fixated ETAs csv file.

    """
    from pharmpy.internals.fs.path import path_absolute
    from pharmpy.model import data

    filename = "fix_eta.csv"
    path = path / filename
    if not force and path.exists():
        raise FileExistsError(f'File at {path} already exists.')

    path = path_absolute(path)
    modelfit_results.individual_estimates.to_csv(path, na_rep=data.conf.na_rep, index=False)
    return path


def verify_param(model1, res1, model2, res2, est=False):
    tol = 0.01

    if est:
        param1 = res1.parameter_estimates
        param2 = res2.parameter_estimates

        passed = []
        failed = []
        for p1 in param1.index:
            if p1 in param2.index:
                p1_value = param1[p1]
                p2_value = param2[p1]
                diff = p1_value - p2_value
                if abs(diff) > tol:
                    failed.append((p1, diff))
                else:
                    passed.append((p1, diff))

    else:
        param1 = model1.parameters
        param2 = model2.parameters

        passed = []
        failed = []
        for p1 in param1:
            if p1.name in param2.names:
                p2 = param2[p1.name]
                diff = p1.init - p2.init
                if abs(diff) > tol:
                    failed.append((p1.name, diff))
                else:
                    passed.append((p1.name, diff))
    return passed, failed


def parse_modelfit_results(model: pharmpy.model.Model, path: Path) -> Union[None, ModelfitResults]:
    """
    Create ModelfitResults object for given model object taken from values saved in executed Rdata file

    Parameters
    ----------
    model : pharmpy.model.Model
        An nlmixr pharmpy model object.
    path : Path
        A path to folder with model and data files.

    Returns
    -------
    Union[None, ModelfitResults]
        Either return ModelfitResult object or None if Rdata file not found.

    """
    rdata_path = path / (model.name + '.RDATA')
    with warnings.catch_warnings():
        # Supress a numpy deprecation warning
        warnings.simplefilter("ignore")
        import pyreadr
    try:
        rdata = pyreadr.read_r(rdata_path)
    except (FileNotFoundError, OSError):
        return None

    rdata["thetas"] = rdata["thetas"].loc[get_thetas(model).names]
    s = []
    for sigma in get_sigmas(model):
        if sigma.init != 1 and not sigma.fix:
            s.append(sigma.name)
    rdata["sigma"] = rdata["sigma"].loc[s]

    ofv = rdata['ofv']['ofv'][0]
    omegas_sigmas = {}
    omega = model.random_variables.etas.covariance_matrix
    for i in range(0, omega.rows):
        for j in range(0, omega.cols):
            symb = omega.row(i)[j]
            if symb != 0:
                omegas_sigmas[symb.name] = rdata['omega'].values[i, j]
    sigma = model.random_variables.epsilons.covariance_matrix
    for i in range(len(sigma)):
        if sigma[i] != 0:
            s = sigma[i]
            if model.parameters[s].init != 1 and not model.parameters[s].fix:
                omegas_sigmas[sigma[i].name] = rdata['sigma']['fit$theta'][sigma[i].name]
    thetas_index = 0
    pe = {}
    for param in model.parameters:
        if param.fix:
            continue
        elif param.name in omegas_sigmas:
            pe[param.name] = omegas_sigmas[param.name]
        else:
            pe[param.name] = rdata['thetas']['fit$theta'][param.name]
            thetas_index += 1

    name = model.name
    description = model.description
    pe = pd.Series(pe)
    predictions = rdata['pred'].set_index(["ID", "TIME"])
    predictions.index = predictions.index.set_levels(
        predictions.index.levels[0].astype("float64"), level=0
    )

    res = ModelfitResults(
        name=name,
        description=description,
        ofv=ofv,
        minimization_successful=True,  # FIXME: Parse minimization status
        parameter_estimates=pe,
        predictions=predictions,
        log=Log(),
    )
    return res



from typing import Iterable

from pharmpy.model import Model
from pharmpy.modeling import add_lag_time, remove_lag_time

from ..statement.feature.lagtime import LagTime
from ..statement.feature.symbols import Wildcard
from ..statement.statement import Statement
from .feature import Feature


def features(model: Model, statements: Iterable[Statement]) -> Iterable[Feature]:
    for statement in statements:
        if isinstance(statement, LagTime):
            modes = (
                statement._wildcard if isinstance(statement.modes, Wildcard) else statement.modes
            )
            for mode in modes:
                if mode.name == "ON":
                    yield ('LAGTIME', mode.name), add_lag_time
                elif mode.name == "OFF":
                    yield ('LAGTIME', mode.name), remove_lag_time
                else:
                    raise ValueError(f'Lagtime {mode} not supported')



import pytest

from pharmpy.model import SimulationStep
from pharmpy.modeling import (
    add_estimation_step,
    add_parameter_uncertainty_step,
    append_estimation_step_options,
    remove_estimation_step,
    remove_parameter_uncertainty_step,
    set_estimation_step,
    set_evaluation_step,
    set_simulation,
)


@pytest.mark.parametrize(
    'method,kwargs,code_ref',
    [
        (
            'fo',
            {'interaction': False},
            '$ESTIMATION METHOD=ZERO MAXEVAL=9990 PRINT=2 POSTHOC',
        ),
        (
            'fo',
            {'interaction': True},
            '$ESTIMATION METHOD=ZERO INTER MAXEVAL=9990 PRINT=2 POSTHOC',
        ),
        (
            'fo',
            {'tool_options': {'saddle_reset': 1}},
            '$ESTIMATION METHOD=ZERO INTER MAXEVAL=9990 PRINT=2 SADDLE_RESET=1',
        ),
        (
            'bayes',
            {'interaction': True},
            '$ESTIMATION METHOD=BAYES INTER MAXEVAL=9990 PRINT=2 POSTHOC',
        ),
        (
            'fo',
            {'interaction': False, 'evaluation': True, 'maximum_evaluations': None},
            '$ESTIMATION METHOD=ZERO MAXEVAL=0 PRINT=2 POSTHOC',
        ),
    ],
)
def test_set_estimation_step(testdata, load_model_for_test, method, kwargs, code_ref):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = set_estimation_step(model, method, **kwargs)
    assert model.model_code.split('\n')[-2] == code_ref


def test_set_estimation_step_est_middle(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = set_estimation_step(
        model, 'FOCE', interaction=True, parameter_uncertainty_method='SANDWICH', idx=0
    )
    assert (
        '$ESTIMATION METHOD=COND INTER MAXEVAL=9990 PRINT=2 POSTHOC\n$COVARIANCE UNCONDITIONAL PRINT=E PRECOND=1'
        in model.model_code
    )


def test_add_estimation_step(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    assert len(model.estimation_steps) == 1
    model = add_estimation_step(model, 'fo')
    assert len(model.estimation_steps) == 2
    assert model.model_code.split('\n')[-2] == '$ESTIMATION METHOD=ZERO'

    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    assert len(model.estimation_steps) == 1
    model = add_estimation_step(model, 'fo', evaluation=True)
    assert len(model.estimation_steps) == 2
    assert model.model_code.split('\n')[-2] == '$ESTIMATION METHOD=ZERO MAXEVAL=0'


def test_add_estimation_step_non_int(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = add_estimation_step(model, 'fo', idx=1.0)
    with pytest.raises(TypeError, match='Index must be integer'):
        add_estimation_step(model, 'fo', idx=1.5)


def test_remove_estimation_step(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    assert len(model.estimation_steps) == 1
    model = remove_estimation_step(model, 0)
    assert not model.estimation_steps
    assert model.model_code.split('\n')[-2] == '$SIGMA 1'


def test_add_parameter_uncertainty_step(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    assert len(model.estimation_steps) == 1
    model = add_parameter_uncertainty_step(model, 'SANDWICH')
    assert len(model.estimation_steps) == 1
    assert model.model_code.split('\n')[-2] == '$COVARIANCE UNCONDITIONAL PRINT=E PRECOND=1'
    model = remove_parameter_uncertainty_step(model)
    model = add_parameter_uncertainty_step(model, 'CPG')
    assert len(model.estimation_steps) == 1
    assert (
        model.model_code.split('\n')[-2] == '$COVARIANCE MATRIX=S UNCONDITIONAL PRINT=E PRECOND=1'
    )

    model = remove_parameter_uncertainty_step(model)

    model = add_parameter_uncertainty_step(model, "EFIM")
    assert len(model.estimation_steps) == 1
    assert (
        "$ESTIMATION METHOD=COND INTER MAXEVAL=9990 PRINT=2 POSTHOC MSFO=efim.msf\n"
        "$PROBLEM DESIGN\n"
        "$DATA file.csv IGNORE=@ REWIND\n"
        "$INPUT ID DV TIME\n"
        "$MSFI efim.msf\n"
        "$DESIGN APPROX=FO FIMDIAG=1 GROUPSIZE=1 OFVTYPE=1\n" in model.model_code
    )


def test_remove_parameter_uncertainty_step(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = add_parameter_uncertainty_step(model, 'SANDWICH')
    assert model.model_code.split('\n')[-2] == '$COVARIANCE UNCONDITIONAL PRINT=E PRECOND=1'
    model = remove_parameter_uncertainty_step(model)
    assert (
        model.model_code.split('\n')[-2]
        == '$ESTIMATION METHOD=COND INTER MAXEVAL=9990 PRINT=2 POSTHOC'
    )

    model = add_parameter_uncertainty_step(model, "EFIM")
    assert len(model.estimation_steps) == 1
    assert (
        "$ESTIMATION METHOD=COND INTER MAXEVAL=9990 PRINT=2 POSTHOC MSFO=efim.msf\n"
        "$PROBLEM DESIGN\n"
        "$DATA file.csv IGNORE=@ REWIND\n"
        "$INPUT ID DV TIME\n"
        "$MSFI efim.msf\n"
        "$DESIGN APPROX=FO FIMDIAG=1 GROUPSIZE=1 OFVTYPE=1\n" in model.model_code
    )
    model = remove_parameter_uncertainty_step(model)
    assert (
        model.model_code.split('\n')[-2]
        == '$ESTIMATION METHOD=COND INTER MAXEVAL=9990 PRINT=2 POSTHOC'
    )


def test_append_estimation_step_options(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    assert len(model.estimation_steps) == 1
    model = append_estimation_step_options(model, {'SADDLE_RESET': 1}, 0)
    assert (
        model.model_code.split('\n')[-2]
        == '$ESTIMATION METHOD=COND INTER MAXEVAL=9990 PRINT=2 POSTHOC SADDLE_RESET=1'
    )


def test_set_evaluation_step(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = set_evaluation_step(model)
    assert (
        model.model_code.split('\n')[-2]
        == '$ESTIMATION METHOD=COND INTER MAXEVAL=0 PRINT=2 POSTHOC'
    )


def test_set_simulation(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = set_simulation(model, n=2, seed=1234)
    assert len(model.estimation_steps) == 1
    assert model.estimation_steps[0] == SimulationStep(n=2, seed=1234)
    assert model.model_code.split('\n')[-2] == "$SIMULATION (1234) SUBPROBLEMS=2"



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import pharmpy.model
from pharmpy.deps import sympy
from pharmpy.internals.code_generator import CodeGenerator
from pharmpy.model.external.nlmixr.error_model import convert_eps_to_sigma
from pharmpy.model.external.nlmixr.model import add_evid
from pharmpy.model.external.nlmixr.model_block import add_bio_lag, add_ode, convert_eq
from pharmpy.model.model import ModelInternals
from pharmpy.modeling import (
    drop_columns,
    get_bioavailability,
    get_lag_times,
    get_thetas,
    translate_nmtran_time,
)


@dataclass(frozen=True)
class RxODEModelInternals(ModelInternals):
    src: Optional[str] = None
    path: Optional[Path] = None


class Model(pharmpy.model.Model):
    def __init__(self, **kwargs):
        super().__init__(
            **kwargs,
        )

    def update_source(self):
        cg = CodeGenerator()
        cg.add(f'{self.name} <- rxode2({{')
        cg.indent()
        create_model(cg, self)
        cg.dedent()
        cg.add('})')
        cg.empty_line()
        create_theta(cg, self)
        cg.empty_line()
        create_eta(cg, self)
        cg.empty_line()
        create_sigma(cg, self)
        cg.empty_line()
        create_fit(cg, self)
        code = str(cg).replace("AMT", "amt").replace("TIME", "time")
        path = None
        internals = self.internals.replace(src=code, path=path)
        model = self.replace(internals=internals)
        return model

    @property
    def model_code(self):
        model = self.update_source()
        code = model.internals.src
        assert code is not None
        return code


def convert_model(model, skip_check=False):
    if isinstance(model, Model):
        return model

    rxode_model = Model(
        internals=RxODEModelInternals(),
        parameters=model.parameters,
        random_variables=model.random_variables,
        statements=model.statements,
        dependent_variables=model.dependent_variables,
        estimation_steps=model.estimation_steps,
        filename_extension='.R',
        datainfo=model.datainfo,
        dataset=model.dataset,
        name=model.name,
        description=model.description,
    )

    # Update dataset
    if model.dataset is not None:
        rxode_model = translate_nmtran_time(rxode_model)

        if all(x in rxode_model.dataset.columns for x in ["RATE", "DUR"]):
            rxode_model = drop_columns(rxode_model, ["DUR"])
        rxode_model = rxode_model.replace(
            datainfo=rxode_model.datainfo.replace(path=None),
            dataset=rxode_model.dataset.reset_index(drop=True),
        )

        # Add evid
        rxode_model = add_evid(rxode_model)

    # Check model for warnings regarding data structure or model contents
    from pharmpy.model.external.nlmixr.sanity_checks import check_model

    rxode_model = check_model(rxode_model, skip_error_model_check=skip_check)

    rxode_model.update_source()

    return rxode_model


def create_model(cg, model):
    add_true_statements(model, cg, model.statements.before_odes)

    if model.statements.ode_system:
        add_ode(model, cg)

    # Add bioavailability statements
    if model.statements.ode_system is not None:
        add_bio_lag(model, cg, bio=True)
        add_bio_lag(model, cg, lag=True)

    # Add statements after ODE
    add_true_statements(model, cg, model.statements.after_odes)


def add_true_statements(model, cg, statements):
    comp_name_dict = {}
    for name in model.statements.ode_system.compartment_names:
        comp = model.statements.ode_system.find_compartment(name)
        comp_name_dict[comp.amount] = comp.amount.name
    for s in statements:
        if model.statements.ode_system is not None and (
            s.symbol in get_bioavailability(model).values()
            or s.symbol in get_lag_times(model).values()
        ):
            pass
        else:
            expr = s.expression.subs(comp_name_dict)
            expr = convert_eps_to_sigma(expr, model)
            if expr.is_Piecewise:
                add_piecewise(model, cg, s)
            else:
                cg.add(f'{s.symbol.name} <- {expr}')


def add_piecewise(model: pharmpy.model.Model, cg: CodeGenerator, s):
    expr = s.expression
    first = True
    for value, cond in expr.args:
        if cond is not sympy.S.true:
            if cond.atoms(sympy.Eq):
                cond = convert_eq(cond)
            if first:
                cg.add(f'if ({cond}) {{')
                first = False
            else:
                cg.add(f'}} else if ({cond}) {{')
        else:
            cg.add('} else {')
            if "NEWIND" in [t.name for t in expr.free_symbols] and value == 0:
                largest_value = expr.args[0].expr
                largest_cond = expr.args[0].cond
                for value, cond in expr.args[1:]:
                    if cond is not sympy.S.true:
                        if cond.rhs > largest_cond.rhs:
                            largest_value = value
                            largest_cond = cond
                        elif cond.rhs == largest_cond.rhs:
                            if not isinstance(cond, sympy.LessThan) and isinstance(
                                largest_cond, sympy.LessThan
                            ):
                                largest_value = value
                                largest_cond = cond
                value = largest_value
        cg.indent()
        value = convert_eps_to_sigma(value, model)
        cg.add(f'{s.symbol.name} <- {value}')
        cg.dedent()
    cg.add('}')


def create_theta(cg, model):
    cg.add("thetas <-")
    cg.add("c(")
    thetas = get_thetas(model)
    for n, theta in enumerate(thetas):
        if n != len(thetas) - 1:
            cg.add(f'{theta.name} = {theta.init}, ')
        else:
            cg.add(f'{theta.name} = {theta.init}')
    cg.add(")")


def create_eta(cg, model):
    from pharmpy.model.external.nlmixr.ini import add_eta

    cg.add("omegas = lotri(")
    add_eta(model, cg, as_list=True)
    cg.add(")")


def create_sigma(cg, model):
    cg.add("sigmas <-")
    cg.add("lotri(")
    all_eps = model.random_variables.epsilons
    for n, eps in enumerate(all_eps):
        sigma = eps.variance
        if len(eps.names) == 1:
            name = model.parameters[sigma].name
            init = model.parameters[sigma].init
            if n != len(all_eps) - 1:
                cg.add(f'{name} ~ {init},')
            else:
                cg.add(f'{name} ~ {init}')
        else:
            cg.add(f'{" + ".join([name for name in eps.names])} ~ c(')
            inits = []
            for row in range(sigma.rows):
                for col in range(row + 1):
                    if col == 0 and row != 0:
                        cg.add(f'{", ".join([str(x) for x in inits])},')
                        inits = []
                        inits.append(f'{model.parameters[sigma[row, col].name].init}')
                    else:
                        inits.append(model.parameters[sigma[row, col].name].init)
            cg.add(f'{", ".join([str(x) for x in inits])}')
            if eps != model.random_variables.epsilons[-1]:
                cg.add("),")
            else:
                cg.add(")")
    cg.add(")")


def create_fit(cg, model):
    cg.add(f'fit <- {model.name} %>% rxSolve(thetas, ev, omega = omegas, sigma = sigmas)')
